from abc import abstractmethod
from collections import defaultdict
from collections.abc import Mapping, Sequence, Set
from typing import TYPE_CHECKING

from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.declarative_automation.operands.subset_automation_condition import (
    SubsetAutomationCondition,
)
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._core.loader import LoadingContext
from dagster._record import record

if TYPE_CHECKING:
    from dagster._core.event_api import EventLogRecord
    from dagster._core.storage.dagster_run import RunRecord


@whitelist_for_serdes
@record
class LatestRunExecutedWithRootTargetCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "executed_with_root_target"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:  # pyright: ignore[reportIncompatibleMethodOverride]
        def _filter_fn(run_record: "RunRecord") -> bool:
            if context.key == context.root_context.key:
                # this happens when this is evaluated for a self-dependent asset. in these cases,
                # it does not make sense to consider the asset as having been executed with itself
                # as the partition key of the target is necessarily different than the partition
                # key of the query key
                return False
            asset_selection = run_record.dagster_run.asset_selection or set()
            check_selection = run_record.dagster_run.asset_check_selection or set()
            return context.root_context.key in (asset_selection | check_selection)

        return await context.asset_graph_view.compute_latest_run_matches_subset(
            from_subset=context.candidate_subset, filter_fn=_filter_fn
        )


def _run_tag_filter_fn(
    run_record: "RunRecord",
    tag_keys: Set[str] | None,
    tag_values: Mapping[str, str] | None,
) -> bool:
    if tag_keys and not all(key in run_record.dagster_run.tags for key in tag_keys):
        return False
    if tag_values and not all(
        run_record.dagster_run.tags.get(key) == value for key, value in tag_values.items()
    ):
        return False
    return True


def _get_run_tag_filter_name(
    base_name: str,
    tag_keys: Set[str] | None,
    tag_values: Mapping[str, str] | None,
) -> str:
    props = []
    name = base_name
    if tag_keys is not None:
        tag_key_str = ",".join(sorted(tag_keys))
        props.append(f"tag_keys={{{tag_key_str}}}")
    if tag_values is not None:
        tag_value_str = ",".join([f"{key}:{value}" for key, value in sorted(tag_values.items())])
        props.append(f"tag_values={{{tag_value_str}}}")

    if props:
        name += f"({', '.join(props)})"
    return name


@whitelist_for_serdes
@record
class LatestRunExecutedWithTagsCondition(SubsetAutomationCondition):
    tag_keys: Set[str] | None = None
    tag_values: Mapping[str, str] | None = None

    @property
    def name(self) -> str:
        return _get_run_tag_filter_name("executed_with_tags", self.tag_keys, self.tag_values)

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:  # pyright: ignore[reportIncompatibleMethodOverride]
        return await context.asset_graph_view.compute_latest_run_matches_subset(
            from_subset=context.candidate_subset,
            filter_fn=lambda run_record: _run_tag_filter_fn(
                run_record, self.tag_keys, self.tag_values
            ),
        )


@record
class NewUpdatesWithRunTagsCondition(SubsetAutomationCondition[AssetKey]):
    tag_keys: Set[str] | None = None
    tag_values: Mapping[str, str] | None = None

    @property
    @abstractmethod
    def base_name(self) -> str: ...

    @property
    def name(self) -> str:
        return _get_run_tag_filter_name(self.base_name, self.tag_keys, self.tag_values)

    @abstractmethod
    def match_candidate_runs(
        self, candidate_run_ids: Set[str], matching_run_ids: Set[str]
    ) -> bool: ...

    async def _get_run_records_from_materializations(
        self,
        materializations: Sequence["EventLogRecord"],
        context: LoadingContext,
    ) -> Sequence["RunRecord"]:
        from dagster._core.storage.dagster_run import RunRecord

        run_ids = list({record.run_id for record in materializations if record.run_id})
        if not run_ids:
            return []

        run_records = await RunRecord.gen_many(context, run_ids)
        return [record for record in run_records if record]

    async def compute_subset(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        context: AutomationContext,
    ) -> EntitySubset[AssetKey]:
        if (
            not context.previous_temporal_context
            or not context.previous_temporal_context.last_event_id
        ):
            return context.get_empty_subset()

        new_materializations = context.asset_graph_view.get_inner_queryer_for_back_compat().get_asset_materializations_updated_after_cursor(
            asset_key=context.key,
            after_cursor=context.previous_temporal_context.last_event_id,
        )
        if not new_materializations:
            return context.get_empty_subset()

        run_records = await self._get_run_records_from_materializations(
            new_materializations,
            context.asset_graph_view,
        )

        matching_run_ids = {
            run_record.dagster_run.run_id
            for run_record in run_records
            if _run_tag_filter_fn(run_record, self.tag_keys, self.tag_values)
        }

        partitions_to_run_ids = defaultdict(set)
        for materialization in new_materializations:
            partitions_to_run_ids[materialization.partition_key].add(materialization.run_id)

        matching_partition_keys = set()

        for partition_key, run_ids in partitions_to_run_ids.items():
            if self.match_candidate_runs(run_ids, matching_run_ids):
                matching_partition_keys.add(partition_key)

        return context.asset_graph_view.get_asset_subset_from_asset_partitions(
            key=context.key,
            asset_partitions={
                AssetKeyPartitionKey(context.key, partition_key)
                for partition_key in matching_partition_keys
            },
            validate_existence=True,
        )


@whitelist_for_serdes
@record
class AnyNewUpdateHasRunTagsCondition(NewUpdatesWithRunTagsCondition):
    @property
    def base_name(self) -> str:
        return "any_new_update_has_run_tags"

    def match_candidate_runs(self, candidate_run_ids: Set[str], matching_run_ids: Set[str]) -> bool:
        # at least one candidate run must have matched
        return len(candidate_run_ids & matching_run_ids) > 0


@whitelist_for_serdes
@record
class AllNewUpdatesHaveRunTagsCondition(NewUpdatesWithRunTagsCondition):
    tag_keys: Set[str] | None = None
    tag_values: Mapping[str, str] | None = None

    @property
    def base_name(self) -> str:
        return "all_new_updates_have_run_tags"

    def match_candidate_runs(self, candidate_run_ids: Set[str], matching_run_ids: Set[str]) -> bool:
        # every candidate run must have matched the filters
        return all(run_id in matching_run_ids for run_id in candidate_run_ids)



import importlib
import inspect
import json
import os
import secrets
import shutil
import string
import sys
import textwrap
from collections.abc import Callable
from pathlib import Path

import yaml
from dagster_shared import check
from dagster_shared.utils import environ

from dagster.components.component_scaffolding import scaffold_object
from dagster.components.core.component_tree import ComponentTree
from dagster.components.scaffold.scaffold import ScaffoldFormatOptions

"""Testing utilities for components."""

import tempfile
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any

from dagster._annotations import public
from dagster._core.definitions.definitions_class import Definitions
from dagster._utils import alter_sys_path
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.core.defs_module import CompositeYamlComponent, get_component


def component_defs(
    *,
    component: Component,
    resources: Mapping[str, Any] | None = None,
    context: ComponentLoadContext | None = None,
) -> Definitions:
    """Builds a Definitions object from a Component.


    Args:
        component (Component): The Component to build the Definitions from.
        resources (Optional[Mapping[str, Any]]): A dictionary of resources to pass to the Component.
        context (Optional[ComponentLoadContext]): A ComponentLoadContext to pass to the Component. Defaults to a test context.

    Examples:

        .. code-block:: python

            # SimpleComponent produces an asset "an_asset"
            an_asset = component_defs(component=SimpleComponent()).get_assets_def("an_asset")
            assert an_asset.key == AssetKey("an_asset")

        .. code-block:: python

            # RequiresResoureComponent produces an asset "an_asset" and requires a resource "my_resource"
            an_asset = component_defs(
                component=SimpleComponent(),
                resources={"my_resource": MyResource()},
            ).get_assets_def("an_asset")
            assert an_asset.key == AssetKey("an_asset")
    """
    context = context or ComponentTree.for_test().load_context
    return component.build_defs(context).with_resources(resources)


def get_original_module_name(cls):
    """Returns the fully qualified module name where a class was originally declared.

    Args:
        cls: A class object (not an instance)

    Returns:
        str: The fully qualified module name (e.g., 'foo.bar.SomeClass')

    Raises:
        TypeError: If the input is not a class
        AttributeError: If the class doesn't have proper module information
    """
    if not isinstance(cls, type):
        raise TypeError("Input must be a class, not an instance")

    # Get the module name where the class is defined
    module_name = cls.__module__

    # Get the class name
    class_name = cls.__qualname__  # Use __qualname__ to handle nested classes properly

    if not module_name:
        raise AttributeError(f"Class {cls} has no module name")

    # Combine them to get the fully qualified name
    return f"{module_name}.{class_name}"


def get_underlying_component(context: ComponentLoadContext) -> Component | None:
    """Loads a component from the given context, resolving the underlying component if
    it is a CompositeYamlComponent.
    """
    component = get_component(context)
    if isinstance(component, CompositeYamlComponent):
        assert len(component.components) == 1
        return component.components[0]
    return component


def random_importable_name(length: int = 8) -> str:
    """Generate a random string that's guaranteed to be importable as a Python symbol."""
    # Ensure it starts with a letter (not underscore to avoid special meanings)
    first_char = secrets.choice(string.ascii_letters)

    # Use only letters and digits for the rest (no underscores for cleaner names)
    remaining_chars = string.ascii_letters + string.digits
    remaining = "".join(secrets.choice(remaining_chars) for _ in range(length - 1))

    return "sandbox_module_" + first_char + remaining


@public
@dataclass
class DefsFolderSandbox:
    """A sandbox for testing components.

    This sandbox provides a number of utilities for scaffolding, modifying, and loading components
    from a temporary defs folder. This makes it easy to test components in isolation.
    """

    project_root: Path
    defs_folder_path: Path
    project_name: str

    @contextmanager
    def build_component_tree(self) -> Iterator[ComponentTree]:
        """Builds a ComponentTree from this sandbox's defs folder."""
        with alter_sys_path(to_add=[str(self.project_root / "src")], to_remove=[]):
            module_path = f"{self.project_name}.defs"
            try:
                tree = ComponentTree(
                    defs_module=importlib.import_module(module_path),
                    project_root=self.project_root,
                )
                yield tree

            finally:
                modules_to_remove = [name for name in sys.modules if name.startswith(module_path)]
                for name in modules_to_remove:
                    del sys.modules[name]

    @contextmanager
    def build_all_defs(self) -> Iterator[Definitions]:
        """Builds a Definitions object corresponding to all components in the sandbox.
        This is equivalent to what would be loaded from a real Dagster project with this
        sandbox's defs folder.
        """
        with self.build_component_tree() as tree:
            yield tree.build_defs()

    @public
    @contextmanager
    def load_component_and_build_defs(
        self, defs_path: Path
    ) -> Iterator[tuple[Component, Definitions]]:
        """Loads a Component object at the given path and builds the corresponding Definitions.

        Args:
            defs_path: The path to the component to load.

        Returns:
            A tuple of the Component and Definitions objects.

        Example:

        .. code-block:: python

            with scaffold_defs_sandbox() as sandbox:
                defs_path = sandbox.scaffold_component(component_cls=MyComponent)
                with sandbox.load_component_and_build_defs(defs_path=defs_path) as (
                    component,
                    defs,
                ):
                    assert isinstance(component, MyComponent)
                    assert defs.get_asset_def("my_asset").key == AssetKey("my_asset")
        """
        with self.build_component_tree() as tree:
            component = tree.load_component(defs_path)
            defs = tree.build_defs(defs_path)
            yield component, defs

    @public
    def scaffold_component(
        self,
        component_cls: Any,
        defs_path: Path | str | None = None,
        scaffold_params: dict[str, Any] | None = None,
        scaffold_format: ScaffoldFormatOptions = "yaml",
        defs_yaml_contents: dict[str, Any] | None = None,
    ) -> Path:
        """Scaffolds a component into the defs folder.

        Args:
            component_cls: The component class to scaffold.
            defs_path: The path to the component. (defaults to a random name)
            scaffold_params: The parameters to pass to the scaffolder.
            scaffold_format: The format to use for scaffolding.
            defs_yaml_contents: The body of the component to update the defs.yaml file with.

        Returns:
            The path to the component.

        Example:

        .. code-block:: python

            with scaffold_defs_sandbox() as sandbox:
                defs_path = sandbox.scaffold_component(component_cls=MyComponent)
                assert (defs_path / "defs.yaml").exists()
        """
        defs_path = self.defs_folder_path / (defs_path or random_importable_name())
        typename = get_original_module_name(component_cls)
        scaffold_object(
            path=defs_path,
            typename=typename,
            json_params=json.dumps(scaffold_params) if scaffold_params else None,
            scaffold_format=scaffold_format,
            project_root=self.project_root,
        )
        if defs_yaml_contents:
            (defs_path / "defs.yaml").write_text(yaml.safe_dump(defs_yaml_contents))
        return defs_path

    @contextmanager
    def swap_defs_file(self, defs_path: Path, defs_yaml_contents: dict[str, Any] | None):
        check.invariant(
            defs_path.suffix == ".yaml",
            "Attributes are only supported for yaml components",
        )
        check.invariant(defs_path.exists(), "defs.yaml must exist")

        # no need to override there is no component body
        if defs_yaml_contents is None:
            yield
            return

        temp_dir = Path(tempfile.mkdtemp())
        temp_path = temp_dir / defs_path.name

        try:
            shutil.copy2(defs_path, temp_path)

            defs_path.write_text(yaml.safe_dump(defs_yaml_contents))

            yield

        finally:
            if temp_path.exists():
                defs_path.unlink(missing_ok=True)
                shutil.copy2(temp_path, defs_path)
            shutil.rmtree(temp_dir)

    @contextmanager
    def activate_venv_for_project(self) -> Iterator[None]:
        """Activates the project's venv and ensures the src directory is in sys.path."""
        injected_path = str(self.project_root / "src")
        try:
            sys.path.insert(1, injected_path)
            venv_path = (self.project_root / ".venv").absolute()
            with environ(
                {
                    "VIRTUAL_ENV": str(venv_path),
                    "PATH": os.pathsep.join(
                        [
                            str(venv_path / ("Scripts" if sys.platform == "win32" else "bin")),
                            os.getenv("PATH", ""),
                        ]
                    ),
                }
            ):
                yield
        finally:
            sys.path.remove(injected_path)


@public
@contextmanager
def create_defs_folder_sandbox(
    *,
    project_name: str | None = None,
) -> Iterator[DefsFolderSandbox]:
    """Create a lightweight sandbox to scaffold and instantiate components. Useful
    for those authoring custom components.

    This function creates a temporary project that mimics the ``defs`` folder portion
    of a real Dagster project. It then yields a :py:class:`DefsFolderSandbox` object which can be used to
    scaffold and load components.

    :py:class:`DefsFolderSandbox` has a few properties useful for different types of tests:

    * ``defs_folder_path``: The absolute path to the ``defs`` folder. The user can inspect and load files from scaffolded components, e.g. ``(defs_folder_path / "my_component" / "defs.yaml").exists()``
    * ``project_name``: If not provided, a random name is generated.

    Once the sandbox is created, you can load all definitions using the ``load`` method on :py:class:`DefsFolderSandbox`, or with the ``load_component_at_path`` method.

    This sandbox does not provide complete environmental isolation, but does provide some isolation guarantees
    to do its best to isolate the test from and restore the environment after the test.

    * A file structure like this is created: ``<<temp folder>> / src / <<project_name>> / defs``
    * ``<<temp folder>> / src`` is placed in ``sys.path`` during the loading process
    * Any modules loaded during the process that descend from defs module are evicted from ``sys.modules`` on cleanup.

    Args:
        project_name: Optional name for the project (default: random name).

    Returns:
        Iterator[DefsFolderSandbox]: A context manager that yields a DefsFolderSandbox

    Example:

    .. code-block:: python

        with create_defs_folder_sandbox() as sandbox:
            defs_path = sandbox.scaffold_component(component_cls=MyComponent)
            assert (defs_path / "defs.yaml").exists()
            assert (defs_path / "my_component_config_file.yaml").exists()  # produced by MyComponentScaffolder

        with create_defs_folder_sandbox() as sandbox:
            defs_path = sandbox.scaffold_component(
                component_cls=MyComponent,
                defs_yaml_contents={"type": "MyComponent", "attributes": {"asset_key": "my_asset"}},
            )
            with sandbox.load_component_and_build_defs(defs_path=defs_path) as (component, defs):
                assert isinstance(component, MyComponent)
                assert defs.get_asset_def("my_asset").key == AssetKey("my_asset")

    """
    project_name = project_name or random_importable_name()

    with tempfile.TemporaryDirectory() as project_root_str:
        project_root = Path(project_root_str)
        defs_folder_path = project_root / "src" / project_name / "defs"
        defs_folder_path.mkdir(parents=True, exist_ok=True)

        dg_toml_path = project_root / "dg.toml"
        dg_toml_path.write_text(
            textwrap.dedent(f"""
                directory_type = "project"
                
                [project]
                root_module = "{project_name}"
                registry_modules = ["dagster_dbt"]
            """)
        )

        yield DefsFolderSandbox(
            project_root=project_root,
            defs_folder_path=defs_folder_path,
            project_name=project_name,
        )


def copy_code_to_file(fn: Callable, file_path: Path) -> None:
    """Takes a function and writes the body of the function to a file.

    Args:
        fn: The function to write to the file.
        file_path: The path to the file to write the function to.

    Example:

    .. code-block:: python

        def code_to_copy() -> None:
            import dagster as dg

            def execute_fn(context) -> dg.MaterializeResult:
                return dg.MaterializeResult()

        copy_code_to_file(code_to_copy, sandbox.defs_folder_path / "execute.py")
    """
    source_code_text = inspect.getsource(fn)
    source_code_text = "\n".join(source_code_text.split("\n")[1:])
    dedented_source_code_text = textwrap.dedent(source_code_text)
    file_path.write_text(dedented_source_code_text)



from collections.abc import Callable

from dagster_shared.merger import deep_merge_dicts

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.partitions.definition import StaticPartitionsDefinition

"""Testing utilities for components."""

from typing import Any, NamedTuple

# Unfortunate hack - we only use this util in pytest tests, we just drop in a no-op
# implementation if pytest is not installed.
try:
    import pytest  # type: ignore
except ImportError:

    class pytest:
        @staticmethod
        def fixture(*args, **kwargs) -> Callable:
            def wrapper(fn):
                return fn

            return wrapper


class TranslationTestCase(NamedTuple):
    name: str
    attributes: dict[str, Any]
    assertion: Callable[[AssetSpec], bool]
    key_modifier: Callable[[AssetKey], AssetKey] | None = None


test_cases = [
    TranslationTestCase(
        name="group_name",
        attributes={"group_name": "group"},
        assertion=lambda asset_spec: asset_spec.group_name == "group",
    ),
    TranslationTestCase(
        name="owners",
        attributes={"owners": ["team:analytics"]},
        assertion=lambda asset_spec: asset_spec.owners == ["team:analytics"],
    ),
    TranslationTestCase(
        name="tags",
        attributes={"tags": {"foo": "bar"}},
        assertion=lambda asset_spec: asset_spec.tags.get("foo") == "bar",
    ),
    TranslationTestCase(
        name="kinds",
        attributes={"kinds": ["snowflake", "dbt"]},
        assertion=lambda asset_spec: "snowflake" in asset_spec.kinds and "dbt" in asset_spec.kinds,
    ),
    TranslationTestCase(
        name="tags-and-kinds",
        attributes={"tags": {"foo": "bar"}, "kinds": ["snowflake", "dbt"]},
        assertion=lambda asset_spec: (
            "snowflake" in asset_spec.kinds
            and "dbt" in asset_spec.kinds
            and asset_spec.tags.get("foo") == "bar"
        ),
    ),
    TranslationTestCase(
        name="code-version",
        attributes={"code_version": "1"},
        assertion=lambda asset_spec: asset_spec.code_version == "1",
    ),
    TranslationTestCase(
        name="description",
        attributes={"description": "some description"},
        assertion=lambda asset_spec: asset_spec.description == "some description",
    ),
    TranslationTestCase(
        name="metadata",
        attributes={"metadata": {"foo": "bar"}},
        assertion=lambda asset_spec: asset_spec.metadata.get("foo") == "bar",
    ),
    TranslationTestCase(
        name="deps",
        attributes={"deps": ["nonexistent"]},
        assertion=lambda asset_spec: (
            len(asset_spec.deps) == 1 and asset_spec.deps[0].asset_key == AssetKey("nonexistent")
        ),
    ),
    TranslationTestCase(
        name="automation_condition",
        attributes={"automation_condition": "{{ automation_condition.eager() }}"},
        assertion=lambda asset_spec: asset_spec.automation_condition is not None,
    ),
    TranslationTestCase(
        name="key",
        attributes={"key": "{{ spec.key.to_user_string() + '_suffix' }}"},
        assertion=lambda asset_spec: asset_spec.key.path[-1].endswith("_suffix"),
        key_modifier=lambda key: AssetKey(path=list(key.path[:-1]) + [f"{key.path[-1]}_suffix"]),
    ),
    TranslationTestCase(
        name="key_prefix",
        attributes={"key_prefix": "cool_prefix"},
        assertion=lambda asset_spec: asset_spec.key.has_prefix(["cool_prefix"]),
        key_modifier=lambda key: AssetKey(path=["cool_prefix"] + list(key.path)),
    ),
    TranslationTestCase(
        name="partitions_defs",
        attributes={"partitions_def": {"type": "static", "partition_keys": ["foo", "bar"]}},
        assertion=lambda asset_spec: isinstance(
            asset_spec.partitions_def, StaticPartitionsDefinition
        ),
    ),
]


class TestTranslation:
    """Pytest test class for testing translation of asset attributes. You can subclass
    this class and implement a test_translation function using the various fixtures in
    order to comprehensively test asset translation options for your component.
    """

    @pytest.fixture(params=test_cases, ids=[case.name for case in test_cases])
    def translation_test_case(self, request):
        return request.param

    @pytest.fixture
    def attributes(self, translation_test_case: TranslationTestCase):
        return translation_test_case.attributes

    @pytest.fixture
    def assertion(self, translation_test_case: TranslationTestCase):
        return translation_test_case.assertion

    @pytest.fixture
    def key_modifier(self, translation_test_case: TranslationTestCase):
        return translation_test_case.key_modifier


class TestTranslationBatched(TestTranslation):
    """This version of the TestTranslation class is used to test the translation of
    asset attributes, applying all customizations in parallel to speed up tests for
    components which might be expensive to construct.
    """

    @pytest.fixture()
    def translation_test_case(self, request):
        deep_merge_all_attributes = {}
        for case in test_cases:
            deep_merge_all_attributes = deep_merge_dicts(deep_merge_all_attributes, case.attributes)
        merged_assertion = lambda asset_spec: all(case.assertion(asset_spec) for case in test_cases)

        # successively apply key modifiers
        def _merged_key_modifier(key):
            for case in test_cases:
                if case.key_modifier:
                    key = case.key_modifier(key)
            return key

        return TranslationTestCase(
            name="merged",
            attributes=deep_merge_all_attributes,
            assertion=merged_assertion,
            key_modifier=_merged_key_modifier,
        )


class TestOpCustomization:
    """Pytest test class for testing customization of op spec. You can subclass
    this class and implement a test_op_customization function using the various fixtures in
    order to comprehensively test op spec customization options for your component.
    """

    @pytest.fixture(
        params=[
            (
                {"name": "my_op"},
                lambda op: op.name == "my_op",
            ),
            (
                {"tags": {"foo": "bar"}},
                lambda op: op.tags.get("foo") == "bar",
            ),
            (
                {"backfill_policy": {"type": "single_run"}},
                lambda op: op.backfill_policy.max_partitions_per_run is None,
            ),
        ],
        ids=["name", "tags", "backfill_policy"],
    )
    def translation_test_case(self, request):
        return request.param

    @pytest.fixture
    def attributes(self, translation_test_case):
        return translation_test_case[0]

    @pytest.fixture
    def assertion(self, translation_test_case):
        return translation_test_case[1]



from collections.abc import Iterator
from typing import AbstractSet, Any, Callable, NamedTuple  # noqa: UP035

import dagster._check as check
from dagster._annotations import PublicAttr, public
from dagster._core.decorator_utils import get_function_params
from dagster._core.definitions.resource_requirement import (
    HookResourceRequirement,
    ResourceRequirement,
)
from dagster._core.definitions.utils import check_valid_name
from dagster._core.errors import DagsterInvalidInvocationError


@public
class HookDefinition(
    NamedTuple(
        "_HookDefinition",
        [
            ("name", PublicAttr[str]),
            ("hook_fn", PublicAttr[Callable]),
            ("required_resource_keys", PublicAttr[AbstractSet[str]]),
            ("decorated_fn", PublicAttr[Callable | None]),
        ],
    ),
):
    """Define a hook which can be triggered during a op execution (e.g. a callback on the step
    execution failure event during a op execution).

    Args:
        name (str): The name of this hook.
        hook_fn (Callable): The callback function that will be triggered.
        required_resource_keys (Optional[AbstractSet[str]]): Keys for the resources required by the
            hook.
    """

    def __new__(
        cls,
        *,
        name: str,
        hook_fn: Callable[..., Any],
        required_resource_keys: AbstractSet[str] | None = None,
        decorated_fn: Callable[..., Any] | None = None,
    ):
        return super().__new__(
            cls,
            name=check_valid_name(name),
            hook_fn=check.callable_param(hook_fn, "hook_fn"),
            required_resource_keys=frozenset(
                check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
            ),
            decorated_fn=check.opt_callable_param(decorated_fn, "decorated_fn"),
        )

    def __call__(self, *args, **kwargs):
        """This is invoked when the hook is used as a decorator.

        We currently support hooks to decorate the following:

        - JobDefinition: when the hook decorates a job definition, it will be added to
            all the op invocations within the job.

        Example:
            .. code-block:: python

                @success_hook
                def slack_message_on_success(_):
                    ...

                @slack_message_on_success
                @job
                def a_job():
                    foo(bar())

        """
        from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
        from dagster._core.definitions.graph_definition import GraphDefinition
        from dagster._core.definitions.hook_invocation import hook_invocation_result
        from dagster._core.definitions.job_definition import JobDefinition
        from dagster._core.execution.context.hook import HookContext

        if len(args) > 0 and isinstance(
            args[0], (JobDefinition, GraphDefinition, AssetsDefinition)
        ):
            # when it decorates a job or asset, we apply this hook to all the op invocations within
            # the job.
            return args[0].with_hooks({self})
        else:
            if not self.decorated_fn:
                raise DagsterInvalidInvocationError(
                    "Only hook definitions created using one of the hook decorators can be invoked."
                )
            fxn_args = get_function_params(self.decorated_fn)
            # If decorated fxn has two arguments, then this is an event list hook fxn, and parameter
            # names are always context and event_list
            if len(fxn_args) == 2:
                context_arg_name = fxn_args[0].name
                event_list_arg_name = fxn_args[1].name
                if len(args) + len(kwargs) != 2:
                    raise DagsterInvalidInvocationError(
                        "Decorated function expects two parameters, context and event_list, but "
                        f"{len(args) + len(kwargs)} were provided."
                    )
                if args:
                    context = check.opt_inst_param(args[0], "context", HookContext)
                    event_list = check.opt_list_param(
                        args[1] if len(args) > 1 else kwargs[event_list_arg_name],
                        event_list_arg_name,
                    )
                else:
                    if context_arg_name not in kwargs:
                        raise DagsterInvalidInvocationError(
                            f"Could not find expected argument '{context_arg_name}'. Provided "
                            f"kwargs: {list(kwargs.keys())}"
                        )
                    if event_list_arg_name not in kwargs:
                        raise DagsterInvalidInvocationError(
                            f"Could not find expected argument '{event_list_arg_name}'. Provided "
                            f"kwargs: {list(kwargs.keys())}"
                        )
                    context = check.opt_inst_param(
                        kwargs[context_arg_name], context_arg_name, HookContext
                    )
                    event_list = check.opt_list_param(
                        kwargs[event_list_arg_name], event_list_arg_name
                    )
                return hook_invocation_result(self, context, event_list)
            else:
                context_arg_name = fxn_args[0].name
                if len(args) + len(kwargs) != 1:
                    raise DagsterInvalidInvocationError(
                        f"Decorated function expects one parameter, {context_arg_name}, but "
                        f"{len(args) + len(kwargs)} were provided."
                    )
                if args:
                    context = check.opt_inst_param(args[0], context_arg_name, HookContext)
                else:
                    if context_arg_name not in kwargs:
                        raise DagsterInvalidInvocationError(
                            f"Could not find expected argument '{context_arg_name}'. Provided "
                            f"kwargs: {list(kwargs.keys())}"
                        )
                    context = check.opt_inst_param(
                        kwargs[context_arg_name], context_arg_name, HookContext
                    )
                return hook_invocation_result(self, context)

    def get_resource_requirements(
        self,
        attached_to: str | None,
    ) -> Iterator[ResourceRequirement]:
        for resource_key in sorted(list(self.required_resource_keys)):
            yield HookResourceRequirement(
                key=resource_key, attached_to=attached_to, hook_name=self.name
            )



from collections.abc import Callable, Iterator
from contextlib import contextmanager
from io import StringIO
from textwrap import TextWrapper
from typing import Any

from typing_extensions import Self

import dagster._check as check

LINE_LENGTH = 100


class IndentingPrinter:
    def __init__(
        self,
        indent_level: int = 2,
        printer: Callable[..., Any] = print,
        current_indent: int = 0,
        line_length: int = LINE_LENGTH,
    ):
        self.current_indent = current_indent
        self.indent_level = check.int_param(indent_level, "indent_level")
        self.printer = check.callable_param(printer, "printer")
        self.line_length = line_length

        self._line_so_far = ""

    def append(self, text: str) -> None:
        check.str_param(text, "text")
        self._line_so_far += text

    def line(self, text: str) -> None:
        check.str_param(text, "text")
        self.printer(self.current_indent_str + self._line_so_far + text)
        self._line_so_far = ""

    def block(self, text: str, prefix: str = "", initial_indent: str = "") -> None:
        """Automagically wrap a block of text."""
        wrapper = TextWrapper(
            width=self.line_length - len(self.current_indent_str),
            initial_indent=initial_indent,
            subsequent_indent=prefix,
            break_long_words=False,
            break_on_hyphens=False,
        )
        for line in wrapper.wrap(text):
            self.line(line)

    def comment(self, text: str) -> None:
        self.block(text, prefix="# ", initial_indent="# ")

    @property
    def current_indent_str(self) -> str:
        return " " * self.current_indent

    def blank_line(self) -> None:
        check.invariant(
            not self._line_so_far, "Cannot throw away appended strings by calling blank_line"
        )
        self.printer("")

    def increase_indent(self) -> None:
        self.current_indent += self.indent_level

    def decrease_indent(self):
        if self.indent_level and self.current_indent <= 0:
            raise Exception("indent cannot be negative")
        self.current_indent -= self.indent_level

    @contextmanager
    def with_indent(self, text: str | None = None) -> Iterator[None]:
        if text is not None:
            self.line(text)
        self.increase_indent()
        yield
        self.decrease_indent()


class IndentingStringIoPrinter(IndentingPrinter):
    """Subclass of IndentingPrinter wrapping a StringIO."""

    def __init__(self, **kwargs: Any):
        self.buffer = StringIO()
        self.printer = lambda x: self.buffer.write(x + "\n")
        super().__init__(printer=self.printer, **kwargs)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, _exception_type, _exception_value, _traceback) -> None:
        self.buffer.close()

    def read(self) -> str:
        """Get the value of the backing StringIO."""
        return self.buffer.getvalue()



import collections.abc
from collections import defaultdict
from collections.abc import Collection, Mapping
from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple, Optional, cast

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.partitions.definition.static import StaticPartitionsDefinition
from dagster._core.definitions.partitions.mapping.partition_mapping import (
    PartitionMapping,
    UpstreamPartitionsResult,
)
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._serdes import whitelist_for_serdes
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


@whitelist_for_serdes
class StaticPartitionMapping(
    PartitionMapping,
    NamedTuple(
        "_StaticPartitionMapping",
        [
            (
                "downstream_partition_keys_by_upstream_partition_key",
                PublicAttr[Mapping[str, str | Collection[str]]],
            )
        ],
    ),
):
    """Define an explicit correspondence between two StaticPartitionsDefinitions.

    Args:
        downstream_partition_keys_by_upstream_partition_key (Dict[str, str | Collection[str]]):
          The single or multi-valued correspondence from upstream keys to downstream keys.
    """

    def __init__(
        self,
        downstream_partition_keys_by_upstream_partition_key: Mapping[str, str | Collection[str]],
    ):
        check.mapping_param(
            downstream_partition_keys_by_upstream_partition_key,
            "downstream_partition_keys_by_upstream_partition_key",
            key_type=str,
            value_type=(str, collections.abc.Collection),
        )

        # cache forward and reverse mappings
        self._mapping = defaultdict(set)
        for (
            upstream_key,
            downstream_keys,
        ) in downstream_partition_keys_by_upstream_partition_key.items():
            self._mapping[upstream_key] = (
                {downstream_keys} if isinstance(downstream_keys, str) else set(downstream_keys)
            )

        self._inverse_mapping = defaultdict(set)
        for upstream_key, downstream_keys in self._mapping.items():
            for downstream_key in downstream_keys:
                self._inverse_mapping[downstream_key].add(upstream_key)

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        if not isinstance(upstream_partitions_def, StaticPartitionsDefinition):
            raise DagsterInvalidDefinitionError(
                "Upstream partitions definition must be a StaticPartitionsDefinition"
            )
        if not isinstance(downstream_partitions_def, StaticPartitionsDefinition):
            raise DagsterInvalidDefinitionError(
                "Downstream partitions definition must be a StaticPartitionsDefinition",
            )
        self._check_upstream(
            upstream_partitions_def=cast("StaticPartitionsDefinition", upstream_partitions_def)
        )
        self._check_downstream(
            downstream_partitions_def=cast("StaticPartitionsDefinition", downstream_partitions_def)
        )

    @cached_method
    def _check_upstream(self, *, upstream_partitions_def: StaticPartitionsDefinition):
        """Validate that the mapping from upstream to downstream is only defined on upstream keys."""
        check.inst_param(
            upstream_partitions_def,
            "upstream_partitions_def",
            StaticPartitionsDefinition,
            "StaticPartitionMapping can only be defined between two StaticPartitionsDefinitions",
        )
        upstream_keys = upstream_partitions_def.get_partition_keys()
        extra_keys = set(self._mapping.keys()).difference(upstream_keys)
        if extra_keys:
            raise ValueError(
                f"mapping source partitions not in the upstream partitions definition: {extra_keys}"
            )

    @cached_method
    def _check_downstream(self, *, downstream_partitions_def: StaticPartitionsDefinition):
        """Validate that the mapping from upstream to downstream only maps to downstream keys."""
        check.inst_param(
            downstream_partitions_def,
            "downstream_partitions_def",
            StaticPartitionsDefinition,
            "StaticPartitionMapping can only be defined between two StaticPartitionsDefinitions",
        )
        downstream_keys = downstream_partitions_def.get_partition_keys()
        extra_keys = set(self._inverse_mapping.keys()).difference(downstream_keys)
        if extra_keys:
            raise ValueError(
                "mapping target partitions not in the downstream partitions definition:"
                f" {extra_keys}"
            )

    def get_downstream_partitions_for_partitions(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: StaticPartitionsDefinition,
        downstream_partitions_def: StaticPartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        with partition_loading_context(current_time, dynamic_partitions_store):
            self._check_downstream(downstream_partitions_def=downstream_partitions_def)

            downstream_subset = downstream_partitions_def.empty_subset()
            downstream_keys = set()
            for key in upstream_partitions_subset.get_partition_keys():
                downstream_keys.update(self._mapping[key])
            return downstream_subset.with_partition_keys(downstream_keys)

    def get_upstream_mapped_partitions_result_for_partitions(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: StaticPartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        with partition_loading_context(current_time, dynamic_partitions_store):
            self._check_upstream(upstream_partitions_def=upstream_partitions_def)

            upstream_subset = upstream_partitions_def.empty_subset()
            if downstream_partitions_subset is None:
                return UpstreamPartitionsResult(
                    partitions_subset=upstream_subset,
                    required_but_nonexistent_subset=upstream_partitions_def.empty_subset(),
                )

            upstream_keys = set()
            for key in downstream_partitions_subset.get_partition_keys():
                upstream_keys.update(self._inverse_mapping[key])

            return UpstreamPartitionsResult(
                partitions_subset=upstream_subset.with_partition_keys(upstream_keys),
                required_but_nonexistent_subset=upstream_partitions_def.empty_subset(),
            )

    @property
    def description(self) -> str:
        return (
            f"Maps upstream partitions to their downstream dependencies according to the "
            f"following mapping: \n{self.downstream_partition_keys_by_upstream_partition_key}"
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Annotated, Any

from dagster_shared import check
from pydantic import BaseModel, ConfigDict, Field

from dagster._annotations import preview, public
from dagster._core.definitions.result import MaterializeResult
from dagster._core.execution.context.asset_check_execution_context import AssetCheckExecutionContext
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster.components.core.context import ComponentLoadContext
from dagster.components.lib.executable_component.component import ExecutableComponent
from dagster.components.lib.sql_component.sql_client import SQLClient
from dagster.components.resolved.core_models import OpSpec
from dagster.components.resolved.model import Resolver


@public
@preview
class SqlComponent(ExecutableComponent, ABC):
    """Base component which executes templated SQL. Subclasses
    implement instructions on where to load the SQL content from.
    """

    # Necessary to allow connection to be a SQLClient, which is an ABC
    model_config = ConfigDict(arbitrary_types_allowed=True)

    connection: Annotated[
        Annotated[
            SQLClient,
            Resolver(lambda ctx, value: value, model_field_type=str),
        ],
        Field(description="The SQL connection to use for executing the SQL content."),
    ]
    execution: Annotated[OpSpec | None, Field(default=None)] = None

    @abstractmethod
    def get_sql_content(
        self, context: AssetExecutionContext, component_load_context: ComponentLoadContext
    ) -> str:
        """The SQL content to execute."""
        ...

    def execute(
        self, context: AssetExecutionContext, component_load_context: ComponentLoadContext
    ) -> None:
        """Execute the SQL content using the Snowflake resource."""
        self.connection.connect_and_execute(self.get_sql_content(context, component_load_context))

    @property
    def op_spec(self) -> OpSpec:
        return self.execution or OpSpec()

    def invoke_execute_fn(
        self,
        context: AssetExecutionContext | AssetCheckExecutionContext,
        component_load_context: ComponentLoadContext,
    ) -> Iterable[MaterializeResult]:
        self.execute(
            check.inst(context, AssetExecutionContext),
            component_load_context,
        )
        for asset in self.assets or []:
            yield MaterializeResult(asset_key=asset.key)


class SqlFile(BaseModel):
    """A file containing SQL content."""

    path: str = Field(..., description="Path to the SQL file")


ResolvedSqlTemplate = Annotated[
    str | SqlFile,
    Resolver(
        lambda ctx, template: template,
        model_field_type=str | SqlFile,
        inject_before_resolve=False,
    ),
]


@public
class TemplatedSqlComponent(SqlComponent):
    """A component which executes templated SQL from a string or file."""

    sql_template: Annotated[
        ResolvedSqlTemplate,
        Field(description="The SQL template to execute, either as a string or from a file."),
    ]

    sql_template_vars: Annotated[
        Mapping[str, Any] | None,
        Field(default=None, description="Template variables to pass to the SQL template."),
    ] = None

    def get_sql_content(
        self, context: AssetExecutionContext, component_load_context: ComponentLoadContext
    ) -> str:
        # Slow import, we do it here to avoid importing jinja2 when `dagster` is imported

        from jinja2 import Template

        template_str = self.sql_template
        if isinstance(template_str, SqlFile):
            template_str = (component_load_context.path / Path(template_str.path)).read_text()

        template = Template(template_str)
        return template.render(**(self.sql_template_vars or {}))



import logging
import os
import sys
import threading
import warnings
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from contextlib import ExitStack
from functools import cached_property
from itertools import count
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Generic,
    Optional,
    TypeAlias,
    TypeVar,
    Union,
)

from typing_extensions import Self

import dagster._check as check
from dagster._config.snap import ConfigTypeSnap
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey
from dagster._core.definitions.assets.graph.remote_asset_graph import (
    RemoteAssetCheckNode,
    RemoteAssetGraph,
    RemoteAssetNode,
    RemoteRepositoryAssetNode,
)
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.data_version import CachingStaleStatusResolver
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.selector import (
    InstigatorSelector,
    JobSelector,
    JobSubsetSelector,
    RepositorySelector,
    ScheduleSelector,
    SensorSelector,
)
from dagster._core.errors import DagsterCodeLocationLoadError, DagsterCodeLocationNotFoundError
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.instance import DagsterInstance
from dagster._core.instance.types import CachingDynamicPartitionsLoader
from dagster._core.loader import LoadingContext
from dagster._core.remote_origin import (
    CodeLocationOrigin,
    GrpcServerCodeLocationOrigin,
    ManagedGrpcPythonEnvCodeLocationOrigin,
)
from dagster._core.remote_representation.code_location import (
    CodeLocation,
    GrpcServerCodeLocation,
    is_implicit_asset_job_name,
)
from dagster._core.remote_representation.external import (
    RemoteExecutionPlan,
    RemoteJob,
    RemotePartitionSet,
    RemoteRepository,
    RemoteSchedule,
    RemoteSensor,
)
from dagster._core.remote_representation.external_data import (
    PartitionConfigSnap,
    PartitionExecutionErrorSnap,
    PartitionNamesSnap,
    PartitionSetExecutionParamSnap,
    PartitionTagsSnap,
    partition_set_snap_name_for_job_name,
)
from dagster._core.remote_representation.grpc_server_registry import GrpcServerRegistry
from dagster._core.remote_representation.grpc_server_state_subscriber import (
    LocationStateChangeEvent,
    LocationStateChangeEventType,
    LocationStateSubscriber,
)
from dagster._core.remote_representation.handle import RepositoryHandle
from dagster._core.snap.dagster_types import DagsterTypeSnap
from dagster._core.snap.mode import ResourceDefSnap
from dagster._core.snap.node import GraphDefSnap, OpDefSnap
from dagster._core.workspace.load_target import WorkspaceLoadTarget
from dagster._core.workspace.permissions import (
    PermissionResult,
    get_location_scoped_user_permissions,
    get_user_permissions,
)
from dagster._core.workspace.workspace import (
    CodeLocationEntry,
    CodeLocationLoadStatus,
    CodeLocationStatusEntry,
    CurrentWorkspace,
    DefinitionsSource,
    location_status_from_location_entry,
)
from dagster._grpc.constants import INCREASE_TIMEOUT_DAGSTER_YAML_MSG, GrpcServerCommand
from dagster._time import get_current_timestamp
from dagster._utils.aiodataloader import DataLoader
from dagster._utils.caching_instance_queryer import CachingInstanceQueryer
from dagster._utils.env import using_dagster_dev
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info

if TYPE_CHECKING:
    from dagster._core.definitions.assets.graph.remote_asset_graph import RemoteWorkspaceAssetGraph
    from dagster._core.remote_representation.external_data import (
        PartitionConfigSnap,
        PartitionExecutionErrorSnap,
        PartitionSetExecutionParamSnap,
    )

T = TypeVar("T")

WEBSERVER_GRPC_SERVER_HEARTBEAT_TTL = 45

RemoteDefinition: TypeAlias = (
    RemoteAssetNode | RemoteAssetCheckNode | RemoteJob | RemoteSchedule | RemoteSensor
)


class BaseWorkspaceRequestContext(LoadingContext):
    """This class is a request-scoped object that stores (1) a reference to all repository locations
    that exist on the `IWorkspaceProcessContext` at the start of the request and (2) a snapshot of the
    workspace at the start of the request.

    This object is needed because a process context and the repository locations on that context can
    be updated (for example, from a thread on the process context). If a request is accessing a
    repository location at the same time the repository location was being cleaned up, we would run
    into errors.
    """

    _exit_stack: ExitStack

    @property
    @abstractmethod
    def instance(self) -> DagsterInstance: ...

    @abstractmethod
    def get_current_workspace(self) -> CurrentWorkspace: ...

    # abstracted since they may be calculated without the full CurrentWorkspace
    @abstractmethod
    def get_location_entry(self, name: str) -> CodeLocationEntry | None: ...

    @abstractmethod
    def get_code_location_statuses(self) -> Sequence[CodeLocationStatusEntry]: ...

    # implemented here since they require the full CurrentWorkspace
    def get_code_location_entries(self) -> Mapping[str, CodeLocationEntry]:
        return self.get_current_workspace().code_location_entries

    def __enter__(self) -> Self:
        self._exit_stack = ExitStack()
        self._exit_stack.enter_context(
            partition_loading_context(dynamic_partitions_store=self.dynamic_partitions_loader)
        )
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self._exit_stack.close()

    @property
    def asset_graph(self) -> "RemoteWorkspaceAssetGraph":
        return self.get_current_workspace().asset_graph

    @cached_property
    def instance_queryer(self) -> CachingInstanceQueryer:
        return CachingInstanceQueryer(
            instance=self.instance,
            asset_graph=self.asset_graph,
            loading_context=self,
        )

    @cached_property
    def dynamic_partitions_loader(self) -> CachingDynamicPartitionsLoader:
        return CachingDynamicPartitionsLoader(self.instance)

    @cached_property
    def stale_status_loader(self) -> CachingStaleStatusResolver:
        return CachingStaleStatusResolver(
            self.instance,
            asset_graph=lambda: self.asset_graph,
            loading_context=self,
        )

    @cached_property
    def data_time_resolver(self) -> CachingDataTimeResolver:
        return CachingDataTimeResolver(self.instance_queryer)

    @property
    @abstractmethod
    def process_context(self) -> "IWorkspaceProcessContext": ...

    @property
    @abstractmethod
    def version(self) -> str | None: ...

    @property
    @abstractmethod
    def permissions(self) -> Mapping[str, PermissionResult]: ...

    @abstractmethod
    def permissions_for_location(self, *, location_name: str) -> Mapping[str, PermissionResult]:
        pass

    @abstractmethod
    def permissions_for_owner(self, *, owner: str) -> Mapping[str, PermissionResult]:
        pass

    def has_permission_for_location(self, permission: str, location_name: str) -> bool:
        if self.has_code_location_name(location_name):
            permissions = self.permissions_for_location(location_name=location_name)
            return permissions[permission].enabled

        # if not in workspace, fall back to the global permissions across all code locations
        return self.has_permission(permission)

    @abstractmethod
    def has_permission(self, permission: str) -> bool: ...

    @abstractmethod
    def was_permission_checked(self, permission: str) -> bool: ...

    def has_permission_for_selector(
        self,
        permission: str,
        selector: AssetKey | AssetCheckKey | JobSelector | ScheduleSelector | SensorSelector,
    ) -> bool:
        if self.has_permission(permission):
            return True

        if isinstance(selector, (AssetKey, AssetCheckKey)):
            if not self.asset_graph.has(selector):
                return False
            location_name = self.asset_graph.get_repository_handle(selector).location_name
        else:
            location_name = selector.location_name

        if not self.has_code_location_name(location_name):
            return False

        if self.has_permission_for_location(permission, location_name):
            return True

        if not self.viewer_has_any_owner_definition_permissions():
            return False

        owners = self.get_owners_for_selector(selector)
        return self.has_permission_for_owners(permission, owners)

    def get_owners_for_selector(
        self,
        selector: AssetKey | AssetCheckKey | JobSelector | ScheduleSelector | SensorSelector,
    ) -> Sequence[str]:
        if isinstance(selector, AssetKey):
            remote_definition = self.asset_graph.get(selector)
        elif isinstance(selector, AssetCheckKey):
            # make asset checks permissioned to the same owners as the underlying asset
            remote_definition = self.asset_graph.get(selector.asset_key)
        elif isinstance(selector, JobSelector):
            remote_definition = self.get_full_job(selector)
        elif isinstance(selector, ScheduleSelector):
            remote_definition = self.get_schedule(selector)
        elif isinstance(selector, SensorSelector):
            remote_definition = self.get_sensor(selector)

        if not remote_definition:
            return []

        return remote_definition.owners or []

    def has_permission_for_owners(self, permission: str, owners: Sequence[str]) -> bool:
        return any(
            self.permissions_for_owner(owner=owner)
            .get(permission, PermissionResult(enabled=False, disabled_reason=None))
            .enabled
            for owner in owners
        )

    @property
    @abstractmethod
    def records_for_run_default_limit(self) -> int | None: ...

    @property
    def show_instance_config(self) -> bool:
        return True

    def viewer_has_any_owner_definition_permissions(self) -> bool:
        return False

    def read_partition_subsets_from_asset_health(self) -> bool:
        return False

    def get_viewer_tags(self) -> dict[str, str]:
        return {}

    def get_reporting_user_tags(self) -> dict[str, str]:
        return {}

    def get_code_location(self, location_name: str) -> CodeLocation:
        location_entry = self.get_location_entry(location_name)
        if not location_entry:
            raise DagsterCodeLocationNotFoundError(
                f"Location {location_name} does not exist in workspace"
            )

        if location_entry.code_location:
            return location_entry.code_location

        if location_entry.load_error:
            error_info = location_entry.load_error
            raise DagsterCodeLocationLoadError(
                f"Failure loading {location_name}: {error_info.to_string()}",
                load_error_infos=[error_info],
            )

        raise DagsterCodeLocationNotFoundError(
            f"Location {location_name} is still loading",
        )

    @property
    def code_locations(self) -> Sequence[CodeLocation]:
        return [
            entry.code_location
            for entry in self.get_code_location_entries().values()
            if entry.code_location
        ]

    @property
    def code_location_names(self) -> Sequence[str]:
        # For some WorkspaceRequestContext subclasses, the CodeLocationEntry is more expensive
        # than the CodeLocationStatusEntry, so use the latter for a faster check.
        return [status_entry.location_name for status_entry in self.get_code_location_statuses()]

    def code_location_errors(self) -> Sequence[SerializableErrorInfo]:
        return [
            entry.load_error
            for entry in self.get_code_location_entries().values()
            if entry.load_error
        ]

    def has_code_location_error(self, name: str) -> bool:
        return self.get_code_location_error(name) is not None

    def get_code_location_error(self, name: str) -> SerializableErrorInfo | None:
        entry = self.get_location_entry(name)
        return entry.load_error if entry else None

    def has_code_location_name(self, name: str) -> bool:
        # For some WorkspaceRequestContext subclasses, the CodeLocationEntry is more expensive
        # than the CodeLocationStatusEntry, so use the latter for a faster check.
        for status_entry in self.get_code_location_statuses():
            if status_entry.location_name == name:
                return True

        return False

    def has_code_location(self, name: str) -> bool:
        location_entry = self.get_location_entry(name)
        return bool(location_entry and location_entry.code_location is not None)

    def is_reload_supported(self, name: str) -> bool:
        entry = self.get_location_entry(name)
        return entry.origin.is_reload_supported if entry else False

    def is_shutdown_supported(self, name: str) -> bool:
        entry = self.get_location_entry(name)
        return entry.origin.is_shutdown_supported if entry else False

    def reload_code_location(self, name: str) -> "BaseWorkspaceRequestContext":
        # This method signals to the remote gRPC server that it should reload its
        # code, and returns a new request context created from the updated process context
        self.process_context.reload_code_location(name)
        return self.process_context.create_request_context()

    def shutdown_code_location(self, name: str):
        self.process_context.shutdown_code_location(name)

    def reload_workspace(self) -> "BaseWorkspaceRequestContext":
        self.process_context.reload_workspace()
        return self.process_context.create_request_context()

    def has_job(self, selector: JobSubsetSelector | JobSelector) -> bool:
        check.inst_param(selector, "selector", (JobSubsetSelector, JobSelector))
        if not self.has_code_location(selector.location_name):
            return False

        loc = self.get_code_location(selector.location_name)
        return loc.has_repository(selector.repository_name) and loc.get_repository(
            selector.repository_name
        ).has_job(selector.job_name)

    def get_full_job(self, selector: JobSubsetSelector | JobSelector) -> RemoteJob:
        return (
            self.get_code_location(selector.location_name)
            .get_repository(selector.repository_name)
            .get_full_job(selector.job_name)
        )

    async def gen_job(
        self,
        selector: JobSubsetSelector,
    ) -> RemoteJob:
        if not selector.is_subset_selection:
            return self.get_full_job(selector)

        return await self.get_code_location(selector.location_name).gen_subset_job(
            selector, self.get_full_job
        )

    def get_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
    ) -> RemoteExecutionPlan:
        return self.get_code_location(remote_job.handle.location_name).get_execution_plan(
            remote_job=remote_job,
            run_config=run_config,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance=self.instance,
        )

    async def gen_execution_plan(
        self,
        remote_job: RemoteJob,
        run_config: Mapping[str, object],
        step_keys_to_execute: Sequence[str] | None,
        known_state: KnownExecutionState | None,
    ) -> RemoteExecutionPlan:
        return await self.get_code_location(remote_job.handle.location_name).gen_execution_plan(
            remote_job=remote_job,
            run_config=run_config,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance=self.instance,
        )

    def get_partition_config(
        self,
        repository_handle: RepositoryHandle,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
    ) -> Union["PartitionConfigSnap", "PartitionExecutionErrorSnap"]:
        return self.get_code_location(repository_handle.location_name).get_partition_config(
            repository_handle=repository_handle,
            job_name=job_name,
            partition_name=partition_name,
            instance=instance,
        )

    def get_partition_tags(
        self,
        repository_selector: RepositorySelector,
        job_name: str,
        partition_name: str,
        instance: DagsterInstance,
        selected_asset_keys: AbstractSet[AssetKey] | None,
    ) -> Union["PartitionTagsSnap", "PartitionExecutionErrorSnap"]:
        if is_implicit_asset_job_name(job_name):
            # Implicit asset jobs never have custom tag-for-partition functions, and the
            # PartitionsDefinitions on the assets are always available on the host, so we can just
            # determine the tags using information on the host.
            # In addition to the performance benefits, this is convenient in the case where the
            # implicit asset job has assets with different PartitionsDefinitions, as the gRPC
            # API for getting partition tags from the code server doesn't support an asset selection.
            partitions_def = self._get_partitions_def_for_job(
                job_selector=JobSelector(
                    location_name=repository_selector.location_name,
                    repository_name=repository_selector.repository_name,
                    job_name=job_name,
                ),
                selected_asset_keys=selected_asset_keys,
            )
            return PartitionTagsSnap(
                name=partition_name,
                tags=check.not_none(partitions_def).get_tags_for_partition_key(partition_name),
            )

        location = self.get_code_location(repository_selector.location_name)
        return location.get_partition_tags_from_repo(
            repository_handle=RepositoryHandle.from_location(
                repository_selector.repository_name,
                location,
            ),
            job_name=job_name,
            partition_name=partition_name,
            instance=instance,
        )

    def get_partition_names(
        self,
        repository_selector: RepositorySelector,
        job_name: str,
        instance: DagsterInstance,
        selected_asset_keys: AbstractSet[AssetKey] | None,
    ) -> Union["PartitionNamesSnap", "PartitionExecutionErrorSnap"]:
        partition_set_name = partition_set_snap_name_for_job_name(job_name)
        partitions_sets = self.get_partition_sets(repository_selector)
        match = next(
            (
                partitions_set
                for partitions_set in partitions_sets
                if partitions_set.name == partition_set_name
            ),
            None,
        )
        if match:
            partition_set = match

            # Prefer to return the names without calling out to user code if there's a corresponding
            # partition set that allows it
            if partition_set.has_partition_name_data():
                return PartitionNamesSnap(
                    partition_names=partition_set.get_partition_names(instance=instance)
                )
            else:
                code_location = self.get_code_location(repository_selector.location_name)
                return code_location.get_partition_names_from_repo(
                    RepositoryHandle.from_location(
                        repository_selector.repository_name,
                        code_location,
                    ),
                    job_name,
                )
        else:
            # Asset jobs might have no corresponding partition set but still have partitioned
            # assets, so we get the partition names using the assets.
            partitions_def = self._get_partitions_def_for_job(
                job_selector=JobSelector(
                    location_name=repository_selector.location_name,
                    repository_name=repository_selector.repository_name,
                    job_name=job_name,
                ),
                selected_asset_keys=selected_asset_keys,
            )
            if not partitions_def:
                return PartitionNamesSnap([])

            return PartitionNamesSnap(
                partitions_def.get_partition_keys(dynamic_partitions_store=instance)
            )

    def _get_partitions_def_for_job(
        self,
        job_selector: JobSelector,
        selected_asset_keys: AbstractSet[AssetKey] | None,
    ) -> PartitionsDefinition | None:
        asset_nodes = self.get_assets_in_job(job_selector, selected_asset_keys)
        unique_partitions_defs: set[PartitionsDefinition] = set()
        for asset_node in asset_nodes:
            if asset_node.asset_node_snap.partitions is not None:
                unique_partitions_defs.add(
                    asset_node.asset_node_snap.partitions.get_partitions_definition()
                )

        if len(unique_partitions_defs) == 0:
            # Assets are all unpartitioned
            return None
        if len(unique_partitions_defs) == 1:
            return next(iter(unique_partitions_defs))
        else:
            check.failed(
                "There is no PartitionsDefinition shared by all the provided assets."
                f" {len(unique_partitions_defs)} unique PartitionsDefinitions."
            )

    def get_partition_set_execution_param_data(
        self,
        repository_handle: RepositoryHandle,
        partition_set_name: str,
        partition_names: Sequence[str],
        instance: DagsterInstance,
    ) -> Union["PartitionSetExecutionParamSnap", "PartitionExecutionErrorSnap"]:
        return self.get_code_location(
            repository_handle.location_name
        ).get_partition_set_execution_params(
            repository_handle=repository_handle,
            partition_set_name=partition_set_name,
            partition_names=partition_names,
            instance=instance,
        )

    def get_notebook_data(self, code_location_name: str, notebook_path: str) -> bytes:
        check.str_param(code_location_name, "code_location_name")
        check.str_param(notebook_path, "notebook_path")
        code_location = self.get_code_location(code_location_name)
        return code_location.get_notebook_data(notebook_path=notebook_path)

    def get_base_deployment_asset_graph(
        self, repository_selector: Optional["RepositorySelector"]
    ) -> Optional["RemoteAssetGraph"]:
        return None

    def get_repository(self, selector: RepositorySelector | RepositoryHandle) -> RemoteRepository:
        return self.get_code_location(selector.location_name).get_repository(
            selector.repository_name
        )

    def get_sensor(self, selector: SensorSelector | InstigatorSelector) -> RemoteSensor | None:
        if not self.has_code_location(selector.location_name):
            return None

        location = self.get_code_location(selector.location_name)

        if not location.has_repository(selector.repository_name):
            return None

        repository = location.get_repository(selector.repository_name)
        if not repository.has_sensor(selector.instigator_name):
            return None

        return repository.get_sensor(selector.instigator_name)

    def get_schedule(
        self, selector: ScheduleSelector | InstigatorSelector
    ) -> RemoteSchedule | None:
        if not self.has_code_location(selector.location_name):
            return None

        location = self.get_code_location(selector.location_name)

        if not location.has_repository(selector.repository_name):
            return None

        repository = location.get_repository(selector.repository_name)
        if not repository.has_schedule(selector.instigator_name):
            return None

        return repository.get_schedule(selector.instigator_name)

    def get_node_def(
        self,
        job_selector: JobSubsetSelector | JobSelector,
        node_def_name: str,
    ) -> OpDefSnap | GraphDefSnap:
        job = self.get_full_job(job_selector)
        return job.get_node_def_snap(node_def_name)

    def get_config_type(
        self,
        job_selector: JobSubsetSelector | JobSelector,
        type_key: str,
    ) -> ConfigTypeSnap:
        job = self.get_full_job(job_selector)
        return job.config_schema_snapshot.get_config_snap(type_key)

    def get_dagster_type(
        self,
        job_selector: JobSubsetSelector | JobSelector,
        type_key: str,
    ) -> DagsterTypeSnap:
        job = self.get_full_job(job_selector)
        return job.job_snapshot.dagster_type_namespace_snapshot.get_dagster_type_snap(type_key)

    def get_resources(
        self,
        job_selector: JobSubsetSelector | JobSelector,
    ) -> Sequence[ResourceDefSnap]:
        job = self.get_full_job(job_selector)
        if not job.mode_def_snaps:
            return []
        return job.mode_def_snaps[0].resource_def_snaps

    def get_dagster_library_versions(self, location_name: str) -> Mapping[str, str] | None:
        return self.get_code_location(location_name).get_dagster_library_versions()

    def get_schedules_targeting_job(
        self,
        selector: JobSubsetSelector | JobSelector,
    ) -> Sequence[RemoteSchedule]:
        repository = self.get_code_location(selector.location_name).get_repository(
            selector.repository_name
        )
        return repository.schedules_by_job_name.get(selector.job_name, [])

    def get_sensors_targeting_job(
        self,
        selector: JobSubsetSelector | JobSelector,
    ) -> Sequence[RemoteSensor]:
        repository = self.get_code_location(selector.location_name).get_repository(
            selector.repository_name
        )
        return repository.sensors_by_job_name.get(selector.job_name, [])

    def get_asset_keys_in_job(
        self,
        selector: JobSubsetSelector | JobSelector,
    ) -> Sequence[AssetKey]:
        if not self.has_code_location(selector.location_name):
            return []

        location = self.get_code_location(selector.location_name)
        if not location.has_repository(selector.repository_name):
            return []

        repository = location.get_repository(selector.repository_name)
        return repository.get_asset_keys_in_job(job_name=selector.job_name)

    def get_assets_in_job(
        self,
        selector: JobSubsetSelector | JobSelector,
        selected_asset_keys: AbstractSet[AssetKey] | None = None,
    ) -> Sequence[RemoteRepositoryAssetNode]:
        keys = self.get_asset_keys_in_job(selector)
        if not keys:
            return []

        if selected_asset_keys is not None:
            keys = [key for key in keys if key in selected_asset_keys]

        repo_asset_graph = self.get_repository(selector.repository_selector).asset_graph
        return [
            repo_asset_graph.get(asset_key) for asset_key in keys if repo_asset_graph.has(asset_key)
        ]

    def get_partition_sets(
        self,
        repository_selector: RepositorySelector,
    ) -> Sequence[RemotePartitionSet]:
        if not self.has_code_location(repository_selector.location_name):
            return []

        location = self.get_code_location(repository_selector.location_name)
        if not location.has_repository(repository_selector.repository_name):
            return []

        repository = location.get_repository(repository_selector.repository_name)
        return repository.get_partition_sets()


class WorkspaceRequestContext(BaseWorkspaceRequestContext):
    def __init__(
        self,
        instance: DagsterInstance,
        current_workspace: CurrentWorkspace,
        process_context: "IWorkspaceProcessContext",
        version: str | None,
        source: object | None,
        read_only: bool,
        read_only_locations: Mapping[str, bool] | None = None,
    ):
        self._instance = instance
        self._current_workspace = current_workspace
        self._process_context = process_context
        self._version = version
        self._source = source
        self._read_only = read_only
        self._read_only_locations = check.opt_mapping_param(
            read_only_locations, "read_only_locations"
        )
        self._checked_permissions: set[str] = set()
        self._loaders = {}

    def reset_for_test(self) -> "WorkspaceRequestContext":
        return WorkspaceRequestContext(
            instance=self.instance,
            current_workspace=self._current_workspace,
            process_context=self._process_context,
            version=self._version,
            source=self._source,
            read_only=self._read_only,
            read_only_locations=self._read_only_locations,
        )

    @property
    def instance(self) -> DagsterInstance:
        return self._instance

    def get_current_workspace(self) -> CurrentWorkspace:
        return self._current_workspace

    def get_location_entry(self, name: str) -> CodeLocationEntry | None:
        return self._current_workspace.code_location_entries.get(name)

    def get_code_location_statuses(self) -> Sequence[CodeLocationStatusEntry]:
        return [
            location_status_from_location_entry(entry)
            for entry in self._current_workspace.code_location_entries.values()
        ]

    @property
    def process_context(self) -> "IWorkspaceProcessContext":
        return self._process_context

    @property
    def version(self) -> str | None:
        return self._version

    @property
    def read_only(self) -> bool:
        return self._read_only

    @property
    def permissions(self) -> Mapping[str, PermissionResult]:
        return get_user_permissions(self._read_only)

    def permissions_for_location(self, *, location_name: str) -> Mapping[str, PermissionResult]:
        if location_name in self._read_only_locations:
            return get_location_scoped_user_permissions(self._read_only_locations[location_name])
        return get_location_scoped_user_permissions(self._read_only)

    def permissions_for_owner(self, *, owner: str) -> Mapping[str, PermissionResult]:
        return {}

    def has_permission(self, permission: str) -> bool:
        permissions = self.permissions
        check.invariant(
            permission in permissions,
            f"Permission {permission} not listed in permissions map",
        )
        self._checked_permissions.add(permission)
        return permissions[permission].enabled

    def has_permission_for_location(self, permission: str, location_name: str) -> bool:
        self._checked_permissions.add(permission)
        return super().has_permission_for_location(permission, location_name)

    def was_permission_checked(self, permission: str) -> bool:
        return permission in self._checked_permissions

    @property
    def source(self) -> object | None:
        """The source of the request this WorkspaceRequestContext originated from.
        For example in the webserver this object represents the web request.
        """
        return self._source

    @property
    def loaders(self) -> dict[type, DataLoader]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self._loaders

    @property
    def records_for_run_default_limit(self) -> int | None:
        return int(os.getenv("DAGSTER_UI_EVENT_LOAD_CHUNK_SIZE", "1000"))


TRequestContext = TypeVar("TRequestContext", bound="BaseWorkspaceRequestContext")


class IWorkspaceProcessContext(ABC, Generic[TRequestContext]):
    """Class that stores process-scoped information about a webserver session.
    In most cases, you will want to create a `BaseWorkspaceRequestContext` to create a request-scoped
    object.
    """

    @abstractmethod
    def create_request_context(self, source: Any | None = None) -> TRequestContext:
        """Create a usable fixed context for the scope of a request.

        Args:
            source (Optional[Any]):
                The source of the request, such as an object representing the web request
                or http connection.
        """

    @property
    @abstractmethod
    def version(self) -> str:
        pass

    @abstractmethod
    def reload_code_location(self, name: str) -> None:
        pass

    def shutdown_code_location(self, name: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def reload_workspace(self) -> None:
        """Reload the code in each code location."""
        pass

    @abstractmethod
    def refresh_workspace(self) -> None:
        """Refresh the snapshots for each code location, without reloading the underlying code."""
        pass

    @property
    @abstractmethod
    def instance(self) -> DagsterInstance:
        pass

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        pass


class WorkspaceProcessContext(IWorkspaceProcessContext[WorkspaceRequestContext]):
    """Process-scoped object that tracks the state of a workspace.

    1. Maintains an update-to-date dictionary of repository locations
    2. Creates a `WorkspaceRequestContext` to be the workspace for each request
    3. Runs watch thread processes that monitor repository locations

    To access a CodeLocation, you should create a `WorkspaceRequestContext`
    using `create_request_context`.
    """

    def __init__(
        self,
        instance: DagsterInstance,
        workspace_load_target: WorkspaceLoadTarget | None,
        version: str = "",
        read_only: bool = False,
        grpc_server_registry: GrpcServerRegistry | None = None,
        code_server_log_level: str = "INFO",
        server_command: GrpcServerCommand = GrpcServerCommand.API_GRPC,
    ):
        self._stack = ExitStack()

        check.opt_str_param(version, "version")
        check.bool_param(read_only, "read_only")

        self._instance = check.inst_param(instance, "instance", DagsterInstance)
        self._workspace_load_target = check.opt_inst_param(
            workspace_load_target, "workspace_load_target", WorkspaceLoadTarget
        )

        self._read_only = read_only

        self._version = version

        # Guards changes to _current_workspace, _watch_thread_shutdown_events and _watch_threads
        self._lock = threading.Lock()
        self._watch_thread_shutdown_events: dict[str, threading.Event] = {}
        self._watch_threads: dict[str, threading.Thread] = {}

        self._state_subscribers_lock = threading.Lock()
        self._state_subscriber_id_iter = count()
        self._state_subscribers: dict[int, LocationStateSubscriber] = {}
        self.add_state_subscriber(LocationStateSubscriber(self._location_state_events_handler))

        if grpc_server_registry:
            self._grpc_server_registry: GrpcServerRegistry = check.inst_param(
                grpc_server_registry, "grpc_server_registry", GrpcServerRegistry
            )
        else:
            self._grpc_server_registry = self._stack.enter_context(
                GrpcServerRegistry(
                    instance_ref=self._instance.get_ref(),
                    server_command=server_command,
                    heartbeat_ttl=WEBSERVER_GRPC_SERVER_HEARTBEAT_TTL,
                    startup_timeout=instance.code_server_process_startup_timeout,
                    log_level=code_server_log_level,
                    wait_for_processes_on_shutdown=instance.wait_for_local_code_server_processes_on_shutdown,
                    additional_timeout_msg=INCREASE_TIMEOUT_DAGSTER_YAML_MSG,
                )
            )

        self._current_workspace: CurrentWorkspace = CurrentWorkspace(code_location_entries={})
        self._update_workspace(
            {
                origin.location_name: self._load_location(origin, reload=False)
                for origin in self._origins
            }
        )

    @property
    def workspace_load_target(self) -> WorkspaceLoadTarget | None:
        return self._workspace_load_target

    @property
    def _origins(self) -> Sequence[CodeLocationOrigin]:
        return self._workspace_load_target.create_origins() if self._workspace_load_target else []

    def get_code_server_specs(self) -> Sequence[Mapping[str, Mapping[str, Any]]]:
        result = []
        for origin in self._origins:
            if isinstance(origin, ManagedGrpcPythonEnvCodeLocationOrigin):
                grpc_endpoint = self._grpc_server_registry.get_grpc_endpoint(origin)
                server_spec = {
                    "location_name": origin.location_name,
                    "socket": grpc_endpoint.socket,
                    "port": grpc_endpoint.port,
                    "host": grpc_endpoint.host,
                    "additional_metadata": origin.loadable_target_origin.as_dict,
                }
            elif isinstance(origin, GrpcServerCodeLocationOrigin):
                server_spec = {
                    "location_name": origin.location_name,
                    "host": origin.host,
                    "port": origin.port,
                    "socket": origin.socket,
                    "additional_metadata": origin.additional_metadata,
                }
            else:
                check.failed(f"Unexpected origin type {origin}")
            result.append({"grpc_server": {k: v for k, v in server_spec.items() if v is not None}})
        return result

    def add_state_subscriber(self, subscriber: LocationStateSubscriber) -> int:
        token = next(self._state_subscriber_id_iter)
        with self._state_subscribers_lock:
            self._state_subscribers[token] = subscriber
        return token

    def rm_state_subscriber(self, token: int) -> None:
        with self._state_subscribers_lock:
            if token in self._state_subscribers:
                del self._state_subscribers[token]

    @property
    def instance(self) -> DagsterInstance:
        return self._instance

    @property
    def read_only(self) -> bool:
        return self._read_only

    @property
    def permissions(self) -> Mapping[str, PermissionResult]:
        return get_user_permissions(True)

    def permissions_for_location(self, *, location_name: str) -> Mapping[str, PermissionResult]:
        return get_location_scoped_user_permissions(True)

    def permissions_for_owner(self, *, owner: str) -> Mapping[str, PermissionResult]:
        return {}

    @property
    def version(self) -> str:
        return self._version

    def _send_state_event_to_subscribers(self, event: LocationStateChangeEvent) -> None:
        check.inst_param(event, "event", LocationStateChangeEvent)
        with self._state_subscribers_lock:
            for subscriber in self._state_subscribers.values():
                subscriber.handle_event(event)

    def _start_watch_thread(self, origin: GrpcServerCodeLocationOrigin) -> None:
        from dagster._grpc.server_watcher import create_grpc_watch_thread

        location_name = origin.location_name
        check.invariant(location_name not in self._watch_thread_shutdown_events)
        client = origin.create_client()
        shutdown_event, watch_thread = create_grpc_watch_thread(
            location_name,
            client,
            on_updated=lambda location_name, new_server_id: self._send_state_event_to_subscribers(
                LocationStateChangeEvent(
                    LocationStateChangeEventType.LOCATION_UPDATED,
                    location_name=location_name,
                    message="Server has been updated.",
                    server_id=new_server_id,
                )
            ),
            on_error=lambda location_name: self._send_state_event_to_subscribers(
                LocationStateChangeEvent(
                    LocationStateChangeEventType.LOCATION_ERROR,
                    location_name=location_name,
                    message=(
                        "Unable to reconnect to server. You can reload the server once it is "
                        "reachable again"
                    ),
                )
            ),
        )
        self._watch_thread_shutdown_events[location_name] = shutdown_event
        self._watch_threads[location_name] = watch_thread
        watch_thread.start()

    def _load_location(self, origin: CodeLocationOrigin, reload: bool) -> CodeLocationEntry:
        location_name = origin.location_name
        location = None
        error = None
        try:
            if isinstance(origin, ManagedGrpcPythonEnvCodeLocationOrigin):
                endpoint = (
                    self._grpc_server_registry.reload_grpc_endpoint(origin)
                    if reload
                    else self._grpc_server_registry.get_grpc_endpoint(origin)
                )
                location = GrpcServerCodeLocation(
                    origin=origin,
                    port=endpoint.port,
                    socket=endpoint.socket,
                    host=endpoint.host,
                    heartbeat=True,
                    watch_server=False,
                    grpc_server_registry=self._grpc_server_registry,
                    instance=self._instance,
                )
            else:
                location = (
                    origin.reload_location(self.instance)
                    if reload
                    else origin.create_location(self.instance)
                )

        except Exception:
            error = serializable_error_info_from_exc_info(sys.exc_info())
            # In dagster dev, the code server process already logs the error, so we don't need to log it again from
            # the workspace process context
            if using_dagster_dev():
                warnings.warn(f"Error loading repository location {location_name}")
            else:
                warnings.warn(
                    f"Error loading repository location {location_name}:{error.to_string()}"
                )

        load_time = get_current_timestamp()
        if isinstance(location, GrpcServerCodeLocation):
            version_key = location.server_id
        else:
            version_key = str(load_time)

        return CodeLocationEntry(
            origin=origin,
            code_location=location,
            load_error=error,
            load_status=CodeLocationLoadStatus.LOADED,
            display_metadata=(
                location.get_display_metadata() if location else origin.get_display_metadata()
            ),
            update_timestamp=load_time,
            version_key=version_key,
            definitions_source=DefinitionsSource.CODE_SERVER,
        )

    def get_current_workspace(self) -> CurrentWorkspace:
        with self._lock:
            return self._current_workspace

    @property
    def code_locations_count(self) -> int:
        with self._lock:
            return len(self._current_workspace.code_location_entries)

    @property
    def code_location_names(self) -> Sequence[str]:
        with self._lock:
            return list(self._current_workspace.code_location_entries)

    def has_code_location(self, location_name: str) -> bool:
        check.str_param(location_name, "location_name")

        with self._lock:
            return (
                location_name in self._current_workspace.code_location_entries
                and self._current_workspace.code_location_entries[location_name].code_location
                is not None
            )

    def has_code_location_error(self, location_name: str) -> bool:
        check.str_param(location_name, "location_name")
        with self._lock:
            return (
                location_name in self._current_workspace.code_location_entries
                and self._current_workspace.code_location_entries[location_name].load_error
                is not None
            )

    def reload_code_location(self, name: str) -> None:
        new_entry = self._load_location(
            self._current_workspace.code_location_entries[name].origin, reload=True
        )
        with self._lock:
            # Relying on GC to clean up the old location once nothing else
            # is referencing it
            self._current_workspace = self._current_workspace.with_code_location(name, new_entry)

    def shutdown_code_location(self, name: str) -> None:
        with self._lock:
            self._current_workspace.code_location_entries[name].origin.shutdown_server()

    def refresh_workspace(self) -> None:
        updated_locations = {
            origin.location_name: self._load_location(origin, reload=False)
            for origin in self._origins
        }
        self._update_workspace(updated_locations)

    def reload_workspace(self) -> None:
        updated_locations = {
            origin.location_name: self._load_location(origin, reload=True)
            for origin in self._origins
        }
        self._update_workspace(updated_locations)

    def _update_workspace(self, new_locations: dict[str, CodeLocationEntry]):
        # minimize lock time by only holding while swapping data old to new
        with self._lock:
            previous_events = self._watch_thread_shutdown_events
            self._watch_thread_shutdown_events = {}

            previous_threads = self._watch_threads
            self._watch_threads = {}

            previous_locations = self._current_workspace.code_location_entries
            self._current_workspace = CurrentWorkspace(code_location_entries=new_locations)

            # start monitoring for new locations
            for entry in new_locations.values():
                if isinstance(entry.origin, GrpcServerCodeLocationOrigin):
                    self._start_watch_thread(entry.origin)

        # clean up previous locations
        for event in previous_events.values():
            event.set()

        for watch_thread in previous_threads.values():
            watch_thread.join()

        for entry in previous_locations.values():
            if entry.code_location:
                entry.code_location.cleanup()

    def create_request_context(self, source: object | None = None) -> WorkspaceRequestContext:
        return WorkspaceRequestContext(
            instance=self._instance,
            current_workspace=self.get_current_workspace(),
            process_context=self,
            version=self.version,
            source=source,
            read_only=self._read_only,
        )

    def _location_state_events_handler(self, event: LocationStateChangeEvent) -> None:
        # If the server was updated or we were not able to reconnect, we immediately reload the
        # location handle
        if event.event_type in (
            LocationStateChangeEventType.LOCATION_UPDATED,
            LocationStateChangeEventType.LOCATION_ERROR,
        ):
            # In case of an updated location, reload the handle to get updated repository data and
            # re-attach a subscriber
            # In case of a location error, just reload the handle in order to update the workspace
            # with the correct error messages
            logging.getLogger("dagster-webserver").info(
                f"Received {event.event_type} event for location {event.location_name}, refreshing"
            )
            self.refresh_code_location(event.location_name)

    def refresh_code_location(self, name: str) -> None:
        # This method reloads the webserver's copy of the code from the remote gRPC server without
        # restarting it, and returns a new request context created from the updated process context
        new_entry = self._load_location(
            self._current_workspace.code_location_entries[name].origin, reload=False
        )
        with self._lock:
            # Relying on GC to clean up the old location once nothing else
            # is referencing it
            self._current_workspace = self._current_workspace.with_code_location(name, new_entry)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self._update_workspace({})  # update to empty to close all current locations
        self._stack.close()

    def copy_for_test_instance(self, instance: DagsterInstance) -> "WorkspaceProcessContext":
        """Make a copy with a different instance, created for tests."""
        return WorkspaceProcessContext(
            instance=instance,
            workspace_load_target=self.workspace_load_target,
            version=self.version,
            read_only=self.read_only,
            grpc_server_registry=self._grpc_server_registry,
        )



from dagster_shared.record import record

from dagster._config import Field, Shape


def get_step_dependency_config_field():
    return Field(
        Shape(
            {"require_upstream_step_success": Field(bool, default_value=True)},
        ),
        is_required=False,
        description="Determines when steps will be executed in relation to their upstream dependencies.",
    )


@record
class StepDependencyConfig:
    require_upstream_step_success: bool

    @staticmethod
    def default() -> "StepDependencyConfig":
        return StepDependencyConfig(require_upstream_step_success=True)

    @staticmethod
    def from_config(config_value: dict[str, bool] | None) -> "StepDependencyConfig":
        return StepDependencyConfig(
            require_upstream_step_success=config_value["require_upstream_step_success"]
            if config_value
            else True,
        )



# Re-export all symbols from submodules for backward compatibility
from dagster._symbol_annotations.annotatable import (
    Annotatable as Annotatable,
    T_Annotatable as T_Annotatable,
    copy_annotations as copy_annotations,
)
from dagster._symbol_annotations.lifecycle import (
    BetaInfo as BetaInfo,
    DeprecatedInfo as DeprecatedInfo,
    PreviewInfo as PreviewInfo,
    SupersededInfo as SupersededInfo,
    attach_deprecation_info_and_wrap as attach_deprecation_info_and_wrap,
    beta as beta,
    beta_param as beta_param,
    deprecated as deprecated,
    deprecated_param as deprecated_param,
    get_beta_info as get_beta_info,
    get_beta_param_info as get_beta_param_info,
    get_beta_params as get_beta_params,
    get_deprecated_info as get_deprecated_info,
    get_deprecated_param_info as get_deprecated_param_info,
    get_deprecated_params as get_deprecated_params,
    get_preview_info as get_preview_info,
    get_superseded_info as get_superseded_info,
    has_beta_params as has_beta_params,
    has_deprecated_params as has_deprecated_params,
    hidden_param as hidden_param,
    is_beta as is_beta,
    is_beta_param as is_beta_param,
    is_deprecated as is_deprecated,
    is_deprecated_param as is_deprecated_param,
    is_preview as is_preview,
    is_superseded as is_superseded,
    only_allow_hidden_params_in_kwargs as only_allow_hidden_params_in_kwargs,
    preview as preview,
    superseded as superseded,
)
from dagster._symbol_annotations.public import (
    PUBLIC as PUBLIC,
    PublicAttr as PublicAttr,
    is_public as is_public,
    public as public,
)



from collections.abc import Mapping
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Annotated

from dagster._record import ImportFrom, record
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.definitions.assets.graph.remote_asset_graph import RemoteWorkspaceAssetGraph
    from dagster._core.remote_origin import CodeLocationOrigin
    from dagster._core.remote_representation.code_location import CodeLocation


# For locations that are loaded asynchronously
class CodeLocationLoadStatus(Enum):
    LOADING = "LOADING"  # Waiting for location to load or update
    LOADED = "LOADED"  # Finished loading (may be an error)


class DefinitionsSource(Enum):
    CODE_SERVER = "CODE_SERVER"
    CONNECTION = "CONNECTION"


@record
class CodeLocationEntry:
    origin: Annotated[
        "CodeLocationOrigin",
        ImportFrom("dagster._core.remote_origin"),
    ]
    code_location: (
        Annotated["CodeLocation", ImportFrom("dagster._core.remote_representation.code_location")]
        | None
    )
    load_error: SerializableErrorInfo | None
    load_status: CodeLocationLoadStatus
    display_metadata: Mapping[str, str]
    update_timestamp: float
    version_key: str
    definitions_source: DefinitionsSource


@record
class CodeLocationStatusEntry:
    """Slimmer version of WorkspaceLocationEntry, containing the minimum set of information required
    to know whether the workspace needs reloading.
    """

    location_name: str
    load_status: CodeLocationLoadStatus
    update_timestamp: float
    version_key: str


@record
class CurrentWorkspace:
    code_location_entries: Mapping[str, CodeLocationEntry]

    @cached_property
    def asset_graph(self) -> "RemoteWorkspaceAssetGraph":
        from dagster._core.definitions.assets.graph.remote_asset_graph import (
            RemoteWorkspaceAssetGraph,
        )

        return RemoteWorkspaceAssetGraph.build(self)

    def with_code_location(self, name: str, entry: CodeLocationEntry) -> "CurrentWorkspace":
        return CurrentWorkspace(code_location_entries={**self.code_location_entries, name: entry})


def location_status_from_location_entry(
    entry: CodeLocationEntry,
) -> CodeLocationStatusEntry:
    return CodeLocationStatusEntry(
        location_name=entry.origin.location_name,
        load_status=entry.load_status,
        update_timestamp=entry.update_timestamp,
        version_key=entry.version_key,
    )



import asyncio
import inspect
import shutil
import tempfile
from abc import abstractmethod
from collections.abc import Awaitable
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from uuid import uuid4

from dagster_shared import check
from dagster_shared.serdes.objects.models.defs_state_info import (
    CODE_SERVER_STATE_VERSION,
    LOCAL_STATE_VERSION,
    DefsStateManagementType,
)
from typing_extensions import Self

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.definitions_load_context import (
    DefinitionsLoadContext,
    DefinitionsLoadType,
)
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster._symbol_annotations.public import public
from dagster._utils.env import using_dagster_dev
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.utils.defs_state import DefsStateConfig
from dagster.components.utils.project_paths import get_local_state_path

if TYPE_CHECKING:
    from pydantic import BaseModel


@public
class StateBackedComponent(Component):
    """Base class for components that depend on external state that needs to be fetched and cached.

    State-backed components are designed for integrations where Dagster definitions depend on
    information from external systems (like APIs or compiled artifacts) rather than just code
    and configuration files. The component framework manages the lifecycle of fetching, storing,
    and loading this state.

    Subclasses must implement:
        - ``write_state_to_path``: Fetches state from external sources and writes it to a local path
        - ``build_defs_from_state``: Builds Dagster definitions from the cached state
        - ``defs_state_config``: Property that returns configuration for state management

    Example:
        .. code-block:: python

            import json
            from dataclasses import dataclass
            from pathlib import Path
            from typing import Optional

            import dagster as dg
            from dagster.components import DefsStateConfig, DefsStateConfigArgs, ResolvedDefsStateConfig

            @dataclass
            class MyStateBackedComponent(dg.StateBackedComponent):
                base_url: str
                defs_state: ResolvedDefsStateConfig = DefsStateConfigArgs.local_filesystem()

                @property
                def defs_state_config(self) -> DefsStateConfig:
                    return DefsStateConfig.from_args(
                        self.defs_state, default_key=f"MyComponent[{self.base_url}]"
                    )

                def write_state_to_path(self, state_path: Path) -> None:
                    # Fetch table metadata from external API
                    response = requests.get(f"{self.base_url}/api/tables")
                    tables = response.json()
                    # Write state to file as JSON
                    state_path.write_text(json.dumps(tables))

                def build_defs_from_state(
                    self, context: dg.ComponentLoadContext, state_path: Optional[Path]
                ) -> dg.Definitions:
                    if state_path is None:
                        return dg.Definitions()

                    # Read cached state
                    tables = json.loads(state_path.read_text())

                    # Create one asset per table found in the state
                    assets = []
                    for table in tables:
                        @dg.asset(key=dg.AssetKey(table["name"]))
                        def table_asset():
                            # Fetch and return the actual table data
                            return fetch_table_data(table["name"])

                        assets.append(table_asset)

                    return dg.Definitions(assets=assets)

        YAML configuration:

        .. code-block:: yaml

            # defs.yaml
            type: my_package.MyStateBackedComponent
            attributes:
              base_url: "{{ env.MY_API_URL }}"
              defs_state:
                management_type: LOCAL_FILESYSTEM
    """

    @classmethod
    def load(cls, attributes: Optional["BaseModel"], context: "ComponentLoadContext") -> Self:
        """Loads the component and marks its defs_state_key on the component tree."""
        loaded_component = super().load(attributes, context)
        context.component_tree.mark_component_defs_state_key(
            defs_state_key=loaded_component.defs_state_config.key,
            component_path=context.component_path,
        )
        return loaded_component

    @property
    @abstractmethod
    def defs_state_config(self) -> DefsStateConfig: ...

    async def _write_state_to_path_async(self, state_path: Path) -> None:
        if inspect.iscoroutinefunction(self.write_state_to_path):
            await self.write_state_to_path(state_path)
        else:
            await asyncio.to_thread(self.write_state_to_path, state_path)

    async def _store_code_server_state(
        self, key: str, state_storage: DefsStateStorage | None
    ) -> str:
        load_context = DefinitionsLoadContext.get()
        check.invariant(
            load_context.load_type == DefinitionsLoadType.INITIALIZATION,
            "Attempted to store LEGACY_CODE_SERVER_SNAPSHOTS state outside of the initialization phase.",
        )
        with tempfile.TemporaryDirectory() as temp_dir:
            state_path = Path(temp_dir) / "state"
            await self._write_state_to_path_async(state_path)
            load_context.add_code_server_defs_state_info(key, state_path.read_text())
            if state_storage:
                state_storage.set_latest_version(key, CODE_SERVER_STATE_VERSION)
            return CODE_SERVER_STATE_VERSION

    async def _store_local_filesystem_state(
        self, key: str, state_storage: DefsStateStorage | None, project_root: Path
    ) -> str:
        state_path = get_local_state_path(key, project_root)
        shutil.rmtree(state_path, ignore_errors=True)
        await self._write_state_to_path_async(state_path)
        if state_storage:
            state_storage.set_latest_version(key, LOCAL_STATE_VERSION)
        return LOCAL_STATE_VERSION

    async def _store_versioned_state_storage_state(
        self, key: str, state_storage: DefsStateStorage
    ) -> str:
        with tempfile.TemporaryDirectory() as temp_dir:
            version = str(uuid4())
            state_path = Path(temp_dir) / "state"
            await self._write_state_to_path_async(state_path)
            state_storage.upload_state_from_path(key, version=version, path=state_path)
            return version

    async def refresh_state(self, project_root: Path) -> str:
        """Rebuilds the state for this component and persists it to the current StateStore.

        Args:
            project_root: The root directory of the project.
        """
        key = self.defs_state_config.key
        state_storage = DefsStateStorage.get()

        if (
            self.defs_state_config.management_type
            == DefsStateManagementType.VERSIONED_STATE_STORAGE
        ):
            if state_storage is None:
                raise DagsterInvalidInvocationError(
                    f"Attempted to refresh state for key {key} with management type {self.defs_state_config.management_type} "
                    "without a StateStorage in context. This is likely the result of an internal framework error."
                )
            return await self._store_versioned_state_storage_state(key, state_storage)
        elif self.defs_state_config.management_type == DefsStateManagementType.LOCAL_FILESYSTEM:
            return await self._store_local_filesystem_state(key, state_storage, project_root)
        elif (
            self.defs_state_config.management_type
            == DefsStateManagementType.LEGACY_CODE_SERVER_SNAPSHOTS
        ):
            check.invariant(
                DefinitionsLoadContext.get().load_type == DefinitionsLoadType.INITIALIZATION,
                "Attempted to refresh `LEGACY_CODE_SERVER_SNAPSHOTS` state explicitly, but this can only happen during code server startup.",
            )
            return await self._store_code_server_state(key, state_storage)
        else:
            raise DagsterInvalidInvocationError(
                f"Invalid state storage location: {self.defs_state_config.management_type}"
            )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        key = self.defs_state_config.key
        defs_load_context = DefinitionsLoadContext.get()
        state_storage = DefsStateStorage.get()

        if defs_load_context.load_type == DefinitionsLoadType.INITIALIZATION:
            if (
                # for code server state management, we always refresh the state
                (
                    self.defs_state_config.management_type
                    == DefsStateManagementType.LEGACY_CODE_SERVER_SNAPSHOTS
                )
                or
                # automatically refresh in local dev unless the user opts out
                (using_dagster_dev() and self.defs_state_config.refresh_if_dev)
            ):
                version = asyncio.run(self.refresh_state(context.project_root))
                defs_load_context.add_defs_state_info(key, version)

        with DefinitionsLoadContext.get().state_path(
            self.defs_state_config, state_storage, context.project_root
        ) as state_path:
            return self.build_defs_from_state(context, state_path=state_path)

    @abstractmethod
    def build_defs_from_state(
        self, context: ComponentLoadContext, state_path: Path | None
    ) -> Definitions:
        """Given a state_path, builds a Definitions object based on the state
        contained at that path.

        Args:
            context (ComponentLoadContext): The context associated with this component load.
            state_path (Optional[Path]): The path to the state file. If no state has been
                previously written to the state store, this will be set to None.

        Returns:
            Definitions: The Definitions object built from the state.
        """

    @abstractmethod
    def write_state_to_path(self, state_path: Path) -> None | Awaitable[None]:
        """Fetches and writes required state to a local file. This method can be
        implemented as either sync or async.

        Args:
            state_path (Path): The path to the state file.
        """
        raise NotImplementedError()



import itertools
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple, Optional, cast

import dagster._check as check
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import (
    MultiPartitionsDefinition,
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.mapping.partition_mapping import (
    PartitionMapping,
    UpstreamPartitionsResult,
)
from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._core.definitions.partitions.utils.multi import MultiPartitionKey
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


class DimensionDependency(NamedTuple):
    partition_mapping: PartitionMapping
    upstream_dimension_name: str | None = None
    downstream_dimension_name: str | None = None


class BaseMultiPartitionMapping(ABC):
    @abstractmethod
    def get_dimension_dependencies(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
    ) -> Sequence[DimensionDependency]: ...

    def get_partitions_def(
        self, partitions_def: PartitionsDefinition, dimension_name: str | None
    ) -> PartitionsDefinition:
        if isinstance(partitions_def, MultiPartitionsDefinition):
            if not isinstance(dimension_name, str):
                check.failed("Expected dimension_name to be a string")
            return partitions_def.get_partitions_def_for_dimension(dimension_name)
        return partitions_def

    def _get_dependency_partitions_subset(
        self,
        a_partitions_def: PartitionsDefinition,
        a_partitions_subset: PartitionsSubset,
        b_partitions_def: PartitionsDefinition,
        a_upstream_of_b: bool,
    ) -> UpstreamPartitionsResult | PartitionsSubset:
        """Given two partitions definitions a_partitions_def and b_partitions_def that have a dependency
        relationship (a_upstream_of_b is True if a_partitions_def is upstream of b_partitions_def),
        and a_partition_keys, a list of partition keys in a_partitions_def, returns a list of
        partition keys in the partitions definition b_partitions_def that are
        dependencies of the partition keys in a_partition_keys.
        """
        a_partition_keys_by_dimension = defaultdict(set)
        if isinstance(a_partitions_def, MultiPartitionsDefinition):
            for partition_key in a_partitions_subset.get_partition_keys():
                key = a_partitions_def.get_partition_key_from_str(partition_key)
                for dimension_name, key in key.keys_by_dimension.items():
                    a_partition_keys_by_dimension[dimension_name].add(key)
        else:
            for partition_key in a_partitions_subset.get_partition_keys():
                a_partition_keys_by_dimension[None].add(partition_key)

        # Maps the dimension name and key of a partition in a_partitions_def to the list of
        # partition keys in b_partitions_def that are dependencies of that partition
        dep_b_keys_by_a_dim_and_key: dict[str | None, dict[str | None, list[str]]] = defaultdict(
            lambda: defaultdict(list)
        )
        required_but_nonexistent_upstream_partitions = set()

        b_dimension_partitions_def_by_name: dict[str | None, PartitionsDefinition] = (
            {
                dimension.name: dimension.partitions_def
                for dimension in b_partitions_def.partitions_defs
            }
            if isinstance(b_partitions_def, MultiPartitionsDefinition)
            else {None: b_partitions_def}
        )

        if a_upstream_of_b:
            # a_partitions_def is upstream of b_partitions_def, so we need to map the
            # dimension names of a_partitions_def to the corresponding dependent dimensions of
            # b_partitions_def
            a_dim_to_dependency_b_dim = {
                dimension_mapping.upstream_dimension_name: (
                    dimension_mapping.downstream_dimension_name,
                    dimension_mapping.partition_mapping,
                )
                for dimension_mapping in self.get_dimension_dependencies(
                    a_partitions_def, b_partitions_def
                )
            }

            for a_dim_name, keys in a_partition_keys_by_dimension.items():
                if a_dim_name in a_dim_to_dependency_b_dim:
                    (
                        b_dim_name,
                        dimension_mapping,
                    ) = a_dim_to_dependency_b_dim[a_dim_name]
                    a_dimension_partitions_def = self.get_partitions_def(
                        a_partitions_def, a_dim_name
                    )
                    b_dimension_partitions_def = self.get_partitions_def(
                        b_partitions_def, b_dim_name
                    )
                    for key in keys:
                        # if downstream dimension mapping exists, for a given key, get the list of
                        # downstream partition keys that are dependencies of that key
                        dep_b_keys_by_a_dim_and_key[a_dim_name][key] = list(
                            dimension_mapping.get_downstream_partitions_for_partitions(
                                a_dimension_partitions_def.empty_subset().with_partition_keys(
                                    [key]
                                ),
                                a_dimension_partitions_def,
                                b_dimension_partitions_def,
                            ).get_partition_keys()
                        )

        else:
            # a_partitions_def is downstream of b_partitions_def, so we need to map the
            # dimension names of a_partitions_def to the corresponding dependency dimensions of
            # b_partitions_def
            a_dim_to_dependency_b_dim = {
                dimension_mapping.downstream_dimension_name: (
                    dimension_mapping.upstream_dimension_name,
                    dimension_mapping.partition_mapping,
                )
                for dimension_mapping in self.get_dimension_dependencies(
                    b_partitions_def, a_partitions_def
                )
            }

            for a_dim_name, keys in a_partition_keys_by_dimension.items():
                if a_dim_name in a_dim_to_dependency_b_dim:
                    (
                        b_dim_name,
                        partition_mapping,
                    ) = a_dim_to_dependency_b_dim[a_dim_name]
                    a_dimension_partitions_def = self.get_partitions_def(
                        a_partitions_def, a_dim_name
                    )
                    b_dimension_partitions_def = self.get_partitions_def(
                        b_partitions_def, b_dim_name
                    )
                    for key in keys:
                        mapped_partitions_result = (
                            partition_mapping.get_upstream_mapped_partitions_result_for_partitions(
                                a_dimension_partitions_def.empty_subset().with_partition_keys(
                                    [key]
                                ),
                                a_dimension_partitions_def,
                                b_dimension_partitions_def,
                            )
                        )
                        dep_b_keys_by_a_dim_and_key[a_dim_name][key] = list(
                            mapped_partitions_result.partitions_subset.get_partition_keys()
                        )

                        # enumerating partition keys since the two subsets might be from different
                        # asset keys
                        required_but_nonexistent_upstream_partitions.update(
                            set(
                                mapped_partitions_result.required_but_nonexistent_subset.get_partition_keys()
                            )
                        )

        b_partition_keys = set()

        mapped_a_dim_names = a_dim_to_dependency_b_dim.keys()
        mapped_b_dim_names = [mapping[0] for mapping in a_dim_to_dependency_b_dim.values()]
        unmapped_b_dim_names = list(
            set(b_dimension_partitions_def_by_name.keys()) - set(mapped_b_dim_names)
        )

        for key in a_partitions_subset.get_partition_keys():
            for b_key_values in itertools.product(
                *(
                    [
                        dep_b_keys_by_a_dim_and_key[dim_name][
                            (
                                cast("MultiPartitionsDefinition", a_partitions_def)
                                .get_partition_key_from_str(key)
                                .keys_by_dimension[dim_name]
                                if dim_name
                                else key
                            )
                        ]
                        for dim_name in mapped_a_dim_names
                    ]
                ),
                *[
                    b_dimension_partitions_def_by_name[dim_name].get_partition_keys()
                    for dim_name in unmapped_b_dim_names
                ],
            ):
                b_partition_keys.add(
                    MultiPartitionKey(
                        {
                            cast("str", (mapped_b_dim_names + unmapped_b_dim_names)[i]): key
                            for i, key in enumerate(b_key_values)
                        }
                    )
                    if len(b_key_values) > 1
                    else b_key_values[0]  # type: ignore
                )

        mapped_subset = b_partitions_def.empty_subset().with_partition_keys(b_partition_keys)
        if a_upstream_of_b:
            return mapped_subset
        else:
            return UpstreamPartitionsResult(
                partitions_subset=mapped_subset,
                required_but_nonexistent_subset=DefaultPartitionsSubset(
                    required_but_nonexistent_upstream_partitions
                ),
            )

    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        with partition_loading_context(current_time, dynamic_partitions_store):
            if downstream_partitions_subset is None:
                check.failed("downstream asset is not partitioned")

            result = self._get_dependency_partitions_subset(
                check.not_none(downstream_partitions_def),
                downstream_partitions_subset,
                cast("MultiPartitionsDefinition", upstream_partitions_def),
                a_upstream_of_b=False,
            )

            if not isinstance(result, UpstreamPartitionsResult):
                check.failed("Expected UpstreamPartitionsResult")

            return result

    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        with partition_loading_context(current_time, dynamic_partitions_store):
            if upstream_partitions_subset is None:
                check.failed("upstream asset is not partitioned")

            result = self._get_dependency_partitions_subset(
                upstream_partitions_def,
                upstream_partitions_subset,
                cast("MultiPartitionsDefinition", downstream_partitions_def),
                a_upstream_of_b=True,
            )

            if isinstance(result, UpstreamPartitionsResult):
                check.failed("Expected PartitionsSubset")

            return result


@whitelist_for_serdes
class DimensionPartitionMapping(
    NamedTuple(
        "_DimensionPartitionMapping",
        [
            ("dimension_name", str),
            ("partition_mapping", PartitionMapping),
        ],
    )
):
    """A helper class for MultiPartitionMapping that defines a partition mapping used to calculate
    the dependent partition keys in the selected downstream MultiPartitions definition dimension.

    Args:
        dimension_name (str): The name of the dimension in the downstream MultiPartitionsDefinition.
        partition_mapping (PartitionMapping): The partition mapping object used to calculate
            the downstream dimension partitions from the upstream dimension partitions and vice versa.
    """

    def __new__(
        cls,
        dimension_name: str,
        partition_mapping: PartitionMapping,
    ):
        return super().__new__(
            cls,
            dimension_name=check.str_param(dimension_name, "dimension_name"),
            partition_mapping=check.inst_param(
                partition_mapping, "partition_mapping", PartitionMapping
            ),
        )



import base64
from collections.abc import Mapping, Sequence
from datetime import datetime
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, NamedTuple

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._record import record
from dagster._serdes import deserialize_value, serialize_value, whitelist_for_serdes
from dagster._time import get_timezone

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.subset.time_window import TimeWindowPartitionsSubset


@whitelist_for_serdes
@record
class TimeWindowCursor:
    start_timestamp: int
    end_timestamp: int
    offset_partition_count: int

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        string_serialized = serialize_value(self)
        return base64.b64encode(bytes(string_serialized, encoding="utf-8")).decode("utf-8")

    @classmethod
    def from_cursor(cls, cursor: str):
        return deserialize_value(base64.b64decode(cursor).decode("utf-8"), cls)


class TimeWindow(NamedTuple):
    """An interval that is closed at the start and open at the end.

    Args:
        start (datetime): A datetime that marks the start of the window.
        end (datetime): A datetime that marks the end of the window.
    """

    start: PublicAttr[datetime]
    end: PublicAttr[datetime]


class PartitionRangeStatus(Enum):
    MATERIALIZING = "MATERIALIZING"
    MATERIALIZED = "MATERIALIZED"
    FAILED = "FAILED"


class PartitionTimeWindowStatus:
    def __init__(self, time_window: TimeWindow, status: PartitionRangeStatus):
        self.time_window = time_window
        self.status = status

    def __repr__(self):
        return f"({self.time_window.start} - {self.time_window.end}): {self.status.value}"

    def __eq__(self, other):
        return (
            isinstance(other, PartitionTimeWindowStatus)
            and self.time_window == other.time_window
            and self.status == other.status
        )

    __hash__ = None  # pyright: ignore[reportAssignmentType]


@whitelist_for_serdes(
    storage_name="TimeWindow",  # For back-compat with existing serdes
)
class PersistedTimeWindow(
    NamedTuple(
        "_PersistedTimeWindow", [("start", TimestampWithTimezone), ("end", TimestampWithTimezone)]
    )
):
    """Internal serialized representation of a time interval that is closed at the
    start and open at the end.
    """

    def __new__(
        cls,
        start: TimestampWithTimezone,
        end: TimestampWithTimezone,
    ):
        return super(cls, PersistedTimeWindow).__new__(
            cls,
            start=check.inst_param(start, "start", TimestampWithTimezone),
            end=check.inst_param(end, "end", TimestampWithTimezone),
        )

    @property
    def start_timestamp(self) -> float:
        return self._asdict()["start"].timestamp

    @property
    def start_timestamp_with_timezone(self) -> TimestampWithTimezone:
        return self._asdict()["start"]

    @property
    def end_timestamp_with_timezone(self) -> TimestampWithTimezone:
        return self._asdict()["end"]

    @property
    def end_timestamp(self) -> float:
        return self._asdict()["end"].timestamp

    @cached_property
    def start(self) -> datetime:  # pyright: ignore[reportIncompatibleVariableOverride]
        start_timestamp_with_timezone = self._asdict()["start"]
        return datetime.fromtimestamp(
            start_timestamp_with_timezone.timestamp,
            tz=get_timezone(start_timestamp_with_timezone.timezone),
        )

    @cached_property
    def end(self) -> datetime:  # pyright: ignore[reportIncompatibleVariableOverride]
        end_timestamp_with_timezone = self._asdict()["end"]
        return datetime.fromtimestamp(
            end_timestamp_with_timezone.timestamp,
            tz=get_timezone(end_timestamp_with_timezone.timezone),
        )

    @staticmethod
    def from_public_time_window(tw: TimeWindow, timezone: str):
        return PersistedTimeWindow(
            TimestampWithTimezone(tw.start.timestamp(), timezone),
            TimestampWithTimezone(tw.end.timestamp(), timezone),
        )

    def subtract(self, other: "PersistedTimeWindow") -> Sequence["PersistedTimeWindow"]:
        other_start_timestamp = other.start_timestamp
        start_timestamp = self.start_timestamp
        other_end_timestamp = other.end_timestamp
        end_timestamp = self.end_timestamp

        # Case where the two don't intersect at all - just return self
        # Note that this assumes end is exclusive
        if end_timestamp <= other_start_timestamp or other_end_timestamp <= start_timestamp:
            return [self]

        windows = []

        if other_start_timestamp > start_timestamp:
            windows.append(
                PersistedTimeWindow(start=self._asdict()["start"], end=other._asdict()["start"]),
            )

        if other_end_timestamp < end_timestamp:
            windows.append(
                PersistedTimeWindow(start=other._asdict()["end"], end=self._asdict()["end"])
            )

        return windows

    def to_public_time_window(self) -> TimeWindow:
        """Used for exposing TimeWindows over the public Dagster API."""
        return TimeWindow(start=self.start, end=self.end)


def _flatten(
    high_pri_time_windows: list[PartitionTimeWindowStatus],
    low_pri_time_windows: list[PartitionTimeWindowStatus],
) -> list[PartitionTimeWindowStatus]:
    high_pri_time_windows = sorted(high_pri_time_windows, key=lambda t: t.time_window.start)
    low_pri_time_windows = sorted(low_pri_time_windows, key=lambda t: t.time_window.start)

    high_pri_idx = 0
    low_pri_idx = 0

    filtered_low_pri: list[PartitionTimeWindowStatus] = []

    # slice and dice the low pri time windows so there's no overlap with high pri
    while True:
        if low_pri_idx >= len(low_pri_time_windows):
            # reached end of materialized
            break
        if high_pri_idx >= len(high_pri_time_windows):
            # reached end of failed, add all remaining materialized bc there's no overlap
            filtered_low_pri.extend(low_pri_time_windows[low_pri_idx:])
            break

        low_pri_tw = low_pri_time_windows[low_pri_idx]
        high_pri_tw = high_pri_time_windows[high_pri_idx]

        if low_pri_tw.time_window.start.timestamp() < high_pri_tw.time_window.start.timestamp():
            if low_pri_tw.time_window.end.timestamp() <= high_pri_tw.time_window.start.timestamp():
                # low_pri_tw is entirely before high pri
                filtered_low_pri.append(low_pri_tw)
                low_pri_idx += 1
            else:
                # high pri cuts the low pri short
                filtered_low_pri.append(
                    PartitionTimeWindowStatus(
                        TimeWindow(
                            low_pri_tw.time_window.start,
                            high_pri_tw.time_window.start,
                        ),
                        low_pri_tw.status,
                    )
                )

                if low_pri_tw.time_window.end.timestamp() > high_pri_tw.time_window.end.timestamp():
                    # the low pri time window will continue on the other end of the high pri
                    # and get split in two. Modify low_pri[low_pri_idx] to be
                    # the second half of the low pri time window. It will be added in the next iteration.
                    # (don't add it now, because we need to check if it overlaps with the next high pri)
                    low_pri_time_windows[low_pri_idx] = PartitionTimeWindowStatus(
                        TimeWindow(high_pri_tw.time_window.end, low_pri_tw.time_window.end),
                        low_pri_tw.status,
                    )
                    high_pri_idx += 1
                else:
                    # the rest of the low pri time window is inside the high pri time window
                    low_pri_idx += 1
        else:
            if low_pri_tw.time_window.start.timestamp() >= high_pri_tw.time_window.end.timestamp():
                # high pri is entirely before low pri. The next high pri may overlap
                high_pri_idx += 1
            elif low_pri_tw.time_window.end.timestamp() <= high_pri_tw.time_window.end.timestamp():
                # low pri is entirely within high pri, skip it
                low_pri_idx += 1
            else:
                # high pri cuts out the start of the low pri. It will continue on the other end.
                # Modify low_pri[low_pri_idx] to shorten the start. It will be added
                # in the next iteration. (don't add it now, because we need to check if it overlaps with the next high pri)
                low_pri_time_windows[low_pri_idx] = PartitionTimeWindowStatus(
                    TimeWindow(high_pri_tw.time_window.end, low_pri_tw.time_window.end),
                    low_pri_tw.status,
                )
                high_pri_idx += 1

    # combine the high pri windwos with the filtered low pri windows
    flattened_time_windows = high_pri_time_windows
    flattened_time_windows.extend(filtered_low_pri)
    flattened_time_windows.sort(key=lambda t: t.time_window.start)
    return flattened_time_windows


PARTITION_RANGE_STATUS_PRIORITY = [
    PartitionRangeStatus.MATERIALIZING,
    PartitionRangeStatus.FAILED,
    PartitionRangeStatus.MATERIALIZED,
]


def fetch_flattened_time_window_ranges(
    subsets: Mapping[PartitionRangeStatus, "TimeWindowPartitionsSubset"],
) -> Sequence[PartitionTimeWindowStatus]:
    """Given potentially overlapping subsets, return a flattened list of timewindows where the highest priority status wins
    on overlaps.
    """
    prioritized_subsets = sorted(
        [(status, subset) for status, subset in subsets.items()],
        key=lambda t: PARTITION_RANGE_STATUS_PRIORITY.index(t[0]),
    )

    # progressively add lower priority time windows to the list of higher priority time windows
    flattened_time_window_statuses = []
    for status, subset in prioritized_subsets:
        subset_time_window_statuses = [
            PartitionTimeWindowStatus(tw.to_public_time_window(), status)
            for tw in subset.included_time_windows
        ]
        flattened_time_window_statuses = _flatten(
            flattened_time_window_statuses, subset_time_window_statuses
        )

    return flattened_time_window_statuses



from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Optional, TypeAlias

from dagster_shared.record import record
from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_key import AssetCheckKey
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._core.storage.asset_check_execution_record import (
    AssetCheckExecutionRecordStatus,
    AssetCheckExecutionResolvedStatus,
    AssetCheckPartitionInfo,
)
from dagster._core.storage.dagster_run import FINISHED_STATUSES, DagsterRunStatus, RunsFilter

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance

StatusSubsets: TypeAlias = Mapping[
    AssetCheckExecutionResolvedStatus, SerializableEntitySubset[AssetCheckKey]
]
InProgressRuns: TypeAlias = Mapping[str, SerializableEntitySubset[AssetCheckKey]]


@whitelist_for_serdes
@record
class AssetCheckState(LoadableBy[tuple[AssetCheckKey, PartitionsDefinition | None]]):
    latest_storage_id: int
    subsets: StatusSubsets
    in_progress_runs: InProgressRuns

    def compatible_with(self, partitions_def: PartitionsDefinition | None) -> bool:
        subset = next(iter(self.subsets.values()), None)
        if subset is None:
            return True
        return subset.is_compatible_with_partitions_def(partitions_def)

    @classmethod
    def empty(cls) -> "AssetCheckState":
        return cls(latest_storage_id=0, subsets={}, in_progress_runs={})

    @classmethod
    def _blocking_batch_load(
        cls,
        keys: Iterable[tuple[AssetCheckKey, PartitionsDefinition | None]],
        context: LoadingContext,
    ) -> Iterable[Optional["AssetCheckState"]]:
        keys = list(keys)
        mapping = context.instance.event_log_storage.get_asset_check_state(keys)
        return [mapping.get(check_key) for check_key, _ in keys]

    def with_updates(
        self,
        key: AssetCheckKey,
        partitions_def: PartitionsDefinition | None,
        partition_records: Sequence[AssetCheckPartitionInfo],
        run_statuses: Mapping[str, DagsterRunStatus],
    ) -> "AssetCheckState":
        latest_storage_id = max(
            (
                max(r.latest_check_event_storage_id, r.latest_materialization_storage_id or 0)
                for r in partition_records
            ),
            default=self.latest_storage_id,
        )

        subsets = {
            status: self.subsets.get(status, SerializableEntitySubset.empty(key, partitions_def))
            for status in AssetCheckExecutionResolvedStatus
        }
        in_progress_runs = dict(self.in_progress_runs)

        # update all subsets based on the new partition records
        subsets, in_progress_runs = _process_partition_records(
            key, partitions_def, subsets, in_progress_runs, partition_records
        )
        # then check the run statuses and resolve any previously in-progress runs that have completed
        subsets, in_progress_runs = _process_run_statuses(
            key, partitions_def, subsets, in_progress_runs, run_statuses
        )
        return AssetCheckState(
            latest_storage_id=latest_storage_id,
            subsets=subsets,
            in_progress_runs=in_progress_runs,
        )


def bulk_update_asset_check_state(
    instance: "DagsterInstance",
    keys: Sequence[tuple[AssetCheckKey, PartitionsDefinition | None]],
    initial_states: Mapping[AssetCheckKey, "AssetCheckState"],
) -> Mapping[AssetCheckKey, "AssetCheckState"]:
    check_keys = [key for key, _ in keys]
    partitions_defs_by_key = {key: partitions_def for key, partitions_def in keys}

    # we prefer to do a single fetch for all keys, so we use the minimum storage id of the initial states
    storage_id = min((state.latest_storage_id for state in initial_states.values()), default=0)
    infos = instance.event_log_storage.get_asset_check_partition_info(
        check_keys, after_storage_id=storage_id
    )

    # find the set of run ids we need to fetch to resolve the in-progress runs, and
    # group the partition infos by check key
    run_ids_to_fetch: set[str] = set().union(
        *(state.in_progress_runs.keys() for state in initial_states.values())
    )
    infos_by_key: dict[AssetCheckKey, list[AssetCheckPartitionInfo]] = defaultdict(list)
    for info in infos:
        infos_by_key[info.check_key].append(info)
        if info.latest_execution_status == AssetCheckExecutionRecordStatus.PLANNED:
            run_ids_to_fetch.add(info.latest_planned_run_id)

    # do a bulk fetch for runs across all states
    finished_runs = (
        instance.get_runs(
            filters=RunsFilter(run_ids=list(run_ids_to_fetch), statuses=FINISHED_STATUSES)
        )
        if len(run_ids_to_fetch) > 0
        else []
    )
    finished_runs_status_by_id = {run.run_id: run.status for run in finished_runs}
    return {
        key: initial_states[key].with_updates(
            key, partitions_defs_by_key[key], infos_by_key[key], finished_runs_status_by_id
        )
        for key in check_keys
    }


def _valid_partition_key(
    partition_key: str | None, partitions_def: PartitionsDefinition | None
) -> bool:
    if partitions_def is None:
        return partition_key is None
    else:
        return partition_key is not None and partitions_def.has_partition_key(partition_key)


def _process_partition_records(
    key: AssetCheckKey,
    partitions_def: PartitionsDefinition | None,
    subsets: StatusSubsets,
    in_progress_runs: InProgressRuns,
    partition_infos: Sequence[AssetCheckPartitionInfo],
) -> tuple[StatusSubsets, InProgressRuns]:
    """Returns a set of updated subsets based on new partition records and the latest materialization storage ids."""
    new_subsets = dict(subsets)
    new_in_progress_runs = dict(in_progress_runs)

    for partition_record in partition_infos:
        pk = partition_record.partition_key
        if not _valid_partition_key(pk, partitions_def):
            continue

        partition_subset = SerializableEntitySubset.from_coercible_value(key, pk, partitions_def)

        if partition_record.latest_execution_status == AssetCheckExecutionRecordStatus.PLANNED:
            # Add to IN_PROGRESS and track run
            new_subsets[AssetCheckExecutionResolvedStatus.IN_PROGRESS] = new_subsets[
                AssetCheckExecutionResolvedStatus.IN_PROGRESS
            ].compute_union(partition_subset)
            run_id = partition_record.latest_planned_run_id
            new_in_progress_runs[run_id] = new_in_progress_runs.get(
                run_id, SerializableEntitySubset.empty(key, partitions_def)
            ).compute_union(partition_subset)

        elif partition_record.latest_execution_status in (
            AssetCheckExecutionRecordStatus.SUCCEEDED,
            AssetCheckExecutionRecordStatus.FAILED,
        ):
            if partition_record.is_current:
                # Check is current, set appropriate status
                status = (
                    AssetCheckExecutionResolvedStatus.SUCCEEDED
                    if partition_record.latest_execution_status
                    == AssetCheckExecutionRecordStatus.SUCCEEDED
                    else AssetCheckExecutionResolvedStatus.FAILED
                )
                new_subsets[status] = new_subsets[status].compute_union(partition_subset)
            else:
                # new materialization, clear the check status
                for status in new_subsets:
                    new_subsets[status] = new_subsets[status].compute_difference(partition_subset)

    return new_subsets, new_in_progress_runs


def _process_run_statuses(
    key: AssetCheckKey,
    partitions_def: PartitionsDefinition | None,
    subsets: StatusSubsets,
    in_progress_runs: InProgressRuns,
    run_statuses: Mapping[str, DagsterRunStatus],
) -> tuple[StatusSubsets, InProgressRuns]:
    """Resolve in-progress runs that have completed.

    This checks if any runs tracked in in_progress_runs have finished,
    and moves their partitions to SKIPPED or EXECUTION_FAILED.
    """
    if not in_progress_runs:
        return subsets, in_progress_runs

    delta_skipped, delta_execution_failed, resolved_run_ids = _resolve_in_progress_subsets(
        key, partitions_def, in_progress_runs, run_statuses
    )

    new_in_progress_runs = {
        run_id: subset
        for run_id, subset in in_progress_runs.items()
        if run_id not in resolved_run_ids
    }

    new_subsets = dict(subsets)
    new_subsets[AssetCheckExecutionResolvedStatus.IN_PROGRESS] = (
        new_subsets[AssetCheckExecutionResolvedStatus.IN_PROGRESS]
        .compute_difference(delta_skipped)
        .compute_difference(delta_execution_failed)
    )
    new_subsets[AssetCheckExecutionResolvedStatus.SKIPPED] = new_subsets[
        AssetCheckExecutionResolvedStatus.SKIPPED
    ].compute_union(delta_skipped)
    new_subsets[AssetCheckExecutionResolvedStatus.EXECUTION_FAILED] = new_subsets[
        AssetCheckExecutionResolvedStatus.EXECUTION_FAILED
    ].compute_union(delta_execution_failed)

    return new_subsets, new_in_progress_runs


def _resolve_in_progress_subsets(
    key: AssetCheckKey,
    partitions_def: PartitionsDefinition | None,
    in_progress_runs: InProgressRuns,
    run_statuses: Mapping[str, DagsterRunStatus],
) -> tuple[
    SerializableEntitySubset[AssetCheckKey],
    SerializableEntitySubset[AssetCheckKey],
    set[str],
]:
    """Resolve in-progress runs that have completed.

    Returns:
        Tuple of (delta_skipped, delta_execution_failed, resolved_run_ids)
    """
    empty_subset = SerializableEntitySubset.empty(key, partitions_def)
    delta_skipped = empty_subset
    delta_execution_failed = empty_subset
    resolved_run_ids: set[str] = set()

    for run_id, run_status in run_statuses.items():
        if run_status in FINISHED_STATUSES and run_id in in_progress_runs:
            resolved_run_ids.add(run_id)
            run_subset = in_progress_runs[run_id]
            if run_status == DagsterRunStatus.FAILURE:
                delta_execution_failed = delta_execution_failed.compute_union(run_subset)
            else:
                delta_skipped = delta_skipped.compute_union(run_subset)

    return delta_skipped, delta_execution_failed, resolved_run_ids



import inspect
from collections.abc import Mapping, Sequence
from dataclasses import MISSING, fields, is_dataclass
from enum import Enum, auto
from functools import partial
from types import UnionType
from typing import Annotated, Any, Final, Literal, Optional, TypeVar, Union, get_args, get_origin

import yaml
from dagster_shared.record import get_record_annotations, get_record_defaults, is_record, record
from dagster_shared.utils import safe_is_subclass
from dagster_shared.yaml_utils import try_parse_yaml_with_source_position
from pydantic import BaseModel, PydanticSchemaGenerationError, create_model
from pydantic.fields import Field, FieldInfo

from dagster import _check as check
from dagster._annotations import public
from dagster._utils.pydantic_yaml import _parse_and_populate_model_with_annotated_errors
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.errors import ResolutionException
from dagster.components.resolved.model import Model, Resolver


class _TypeContainer(Enum):
    SEQUENCE = auto()
    OPTIONAL = auto()
    DICT = auto()


_DERIVED_MODEL_REGISTRY = {}


@public
class Resolvable:
    """Base class for making a class resolvable from yaml.

    This framework is designed to allow complex nested objects to be resolved
    from yaml documents. This allows for a single class to be instantiated from
    either yaml or python without limiting the types of fields that can exist on
    the python class.

    Key Features:
    - **Automatic yaml schema derivation**: A pydantic model is automatically generated from the class definition using its fields or `__init__` arguments and their annotations.
    - **Jinja template resolution**: Fields in the yaml document may be templated strings, which are rendered from the available scope and may be arbitrary python objects.
    - **Customizable resolution behavior**: Each field can customize how it is resolved from the yaml document using a `:py:class:~dagster.Resolver`.

    Resolvable subclasses must be one of the following:
    * pydantic model
    * @dataclass
    * plain class with an annotated `__init__`
    * @record

    Example:

    .. code-block:: python

        import datetime
        from typing import Annotated

        import dagster as dg


        def resolve_timestamp(
            context: dg.ResolutionContext,
            raw_timestamp: str,
        ) -> datetime.datetime:
            return datetime.datetime.fromisoformat(
                context.resolve_value(raw_timestamp, as_type=str),
            )


        # the yaml field will be a string, which is then parsed into a datetime object
        ResolvedTimestamp = Annotated[
            datetime.datetime,
            dg.Resolver(resolve_timestamp, model_field_type=str),
        ]


        class MyClass(dg.Resolvable, dg.Model):
            event: str
            start_timestamp: ResolvedTimestamp
            end_timestamp: ResolvedTimestamp


        # python instantiation
        in_python = MyClass(
            event="test",
            start_timestamp=datetime.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            end_timestamp=datetime.datetime(2021, 1, 2, 0, 0, 0, tzinfo=datetime.timezone.utc),
        )

        # yaml instantiation
        in_yaml = MyClass.resolve_from_yaml(
            '''
        event: test
        start_timestamp: '{{ start_year }}-01-01T00:00:00Z'
        end_timestamp: '{{ end_timestamp }}'
        ''',
            scope={
                # string templating
                "start_year": "2021",
                # object templating
                "end_timestamp": in_python.end_timestamp,
            },
        )

        assert in_python == in_yaml

    """

    @classmethod
    def model(cls) -> type[BaseModel]:
        return derive_model_type(cls)

    @classmethod
    def resolve_from_model(cls, context: "ResolutionContext", model: BaseModel):
        return cls(**resolve_fields(model, cls, context))

    @classmethod
    def resolve_from_yaml(
        cls,
        yaml: str,
        *,
        scope: Mapping[str, Any] | None = None,
    ):
        parsed_and_src_tree = try_parse_yaml_with_source_position(yaml)
        model_cls = cls.model()
        if parsed_and_src_tree:
            model = _parse_and_populate_model_with_annotated_errors(
                cls=model_cls,
                obj_parse_root=parsed_and_src_tree,
                obj_key_path_prefix=[],
            )
        else:  # yaml parsed as None
            model = model_cls()

        context = ResolutionContext.default(
            parsed_and_src_tree.source_position_tree if parsed_and_src_tree else None
        )

        if scope:
            context = context.with_scope(**scope)

        return cls.resolve_from_model(context, model)

    @classmethod
    def resolve_from_dict(cls, dictionary: dict[str, Any]):
        # Convert dictionary to YAML string
        # default_flow_style=False makes it use block style instead of inline
        yaml_string = yaml.dump(
            dictionary,
            default_flow_style=False,
            sort_keys=False,  # Preserve dictionary order
            indent=2,  # Set indentation level
        )
        return cls.resolve_from_yaml(yaml_string)


# marker type for skipping kwargs and triggering defaults
# must be a string to make sure it is json serializable
_Unset: Final[str] = "__DAGSTER_UNSET_DEFAULT__"


def derive_model_type(
    target_type: type[Resolvable],
) -> type[BaseModel]:
    if target_type not in _DERIVED_MODEL_REGISTRY:
        model_name = f"{target_type.__name__}Model"

        model_fields: dict[
            str, Any
        ] = {}  # use Any to appease type checker when **-ing in to create_model

        for name, annotation_info in _get_annotations(target_type).items():
            field_resolver = _get_resolver(annotation_info.type, name)
            field_name = field_resolver.model_field_name or name
            field_type = field_resolver.model_field_type or annotation_info.type

            field_infos = []
            if annotation_info.field_info:
                # remove the default_factory as we set a `default` marker field on the Model type
                # for unserializable defaults (and functions are not serializable)
                field_infos.append(
                    FieldInfo.merge_field_infos(annotation_info.field_info, default_factory=None)
                )

            if annotation_info.has_default:
                # if the annotation has a serializable default
                # value, propagate it to the inner schema, otherwise
                # use a marker value that will cause the kwarg
                # to get omitted when we resolve fields in order
                # to trigger the default on the target type
                default_value = (
                    annotation_info.default
                    if type(annotation_info.default) in {int, float, str, bool, type(None)}
                    else _Unset
                )
                field_infos.append(
                    Field(
                        default=default_value,
                        description=field_resolver.description,
                        examples=field_resolver.examples,
                    ),
                )
            elif field_resolver.description or field_resolver.examples:
                field_infos.append(
                    Field(
                        description=field_resolver.description,
                        examples=field_resolver.examples,
                    )
                )

            # make all fields injectable
            if field_type != str:
                field_type = field_type | str

            model_fields[field_name] = (
                field_type,
                FieldInfo.merge_field_infos(*field_infos),
            )

        try:
            _DERIVED_MODEL_REGISTRY[target_type] = create_model(
                model_name,
                __base__=Model,
                **model_fields,
            )
        except PydanticSchemaGenerationError as e:
            raise ResolutionException(f"Unable to derive Model for {target_type}") from e

    return _DERIVED_MODEL_REGISTRY[target_type]


def _is_implicitly_resolved_type(annotation):
    if annotation in (int, float, str, bool, Any, type(None), list, dict):
        return True

    if safe_is_subclass(annotation, Enum):
        return True

    if safe_is_subclass(annotation, Resolvable):
        # ensure valid Resolvable subclass
        annotation.model()
        return False

    if safe_is_subclass(annotation, BaseModel):
        _ensure_non_resolvable_model_compliance(annotation)
        return True

    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin in (Union, UnionType, list, Sequence, tuple, dict, Mapping) and all(
        _is_implicitly_resolved_type(arg) for arg in args
    ):
        return True

    if origin is Literal and all(_is_implicitly_resolved_type(type(arg)) for arg in args):
        return True

    return False


def _is_resolvable_type(annotation):
    return _is_implicitly_resolved_type(annotation) or safe_is_subclass(annotation, Resolvable)


@record
class AnnotationInfo:
    type: Any
    default: Any
    has_default: bool
    field_info: FieldInfo | None


def _get_annotations(
    resolved_type: type[Resolvable],
) -> dict[str, AnnotationInfo]:
    annotations: dict[str, AnnotationInfo] = {}
    init_kwargs = _get_init_kwargs(resolved_type)
    if is_dataclass(resolved_type):
        for f in fields(resolved_type):
            has_default = f.default is not MISSING or f.default_factory is not MISSING
            annotations[f.name] = AnnotationInfo(
                type=f.type,
                default=f.default,
                has_default=has_default,
                field_info=None,
            )
        return annotations
    elif safe_is_subclass(resolved_type, BaseModel):
        for name, field_info in resolved_type.model_fields.items():
            has_default = not field_info.is_required()
            annotations[name] = AnnotationInfo(
                type=field_info.rebuild_annotation(),
                default=field_info.default,
                has_default=has_default,
                field_info=field_info,
            )
        return annotations
    elif is_record(resolved_type):
        defaults = get_record_defaults(resolved_type)
        for name, ttype in get_record_annotations(resolved_type).items():
            annotations[name] = AnnotationInfo(
                type=ttype,
                default=defaults[name] if name in defaults else None,
                has_default=name in defaults,
                field_info=None,
            )
        return annotations
    elif init_kwargs is not None:
        return init_kwargs
    else:
        raise ResolutionException(
            f"Invalid Resolvable type {resolved_type}, could not determine fields. Resolved subclasses must be one of the following:\n"
            "* class with __init__\n"
            "* @dataclass\n"
            "* pydantic Model\n"
            "* @dagster_shared.record.record\n"
        )


def _get_init_kwargs(
    target_type: type[Resolvable],
) -> dict[str, AnnotationInfo] | None:
    if target_type.__init__ is object.__init__:
        return None

    sig = inspect.signature(target_type.__init__)
    fields: dict[str, AnnotationInfo] = {}

    skipped_self = False
    for name, param in sig.parameters.items():
        if not skipped_self:
            skipped_self = True
            continue

        if param.kind == param.POSITIONAL_ONLY:
            raise ResolutionException(
                f"Invalid Resolvable type {target_type}: __init__ contains positional only parameter."
            )
        if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            continue
        if param.annotation == param.empty:
            raise ResolutionException(
                f"Invalid Resolvable type {target_type}: __init__ parameter {name} has no type hint."
            )

        fields[name] = AnnotationInfo(
            type=param.annotation,
            default=param.default,
            has_default=param.default is not param.empty,
            field_info=None,
        )
    return fields


def resolve_fields(
    model: BaseModel,
    resolved_cls: type,
    context: "ResolutionContext",
) -> Mapping[str, Any]:
    """Returns a mapping of field names to resolved values for those fields."""
    alias_name_by_field_name = {
        field_name: (
            annotation_info.field_info.alias
            if annotation_info.field_info and annotation_info.field_info.alias
            else field_name
        )
        for field_name, annotation_info in _get_annotations(resolved_cls).items()
    }
    field_resolvers = {
        (field_name): _get_resolver(annotation_info.type, field_name)
        for field_name, annotation_info in _get_annotations(resolved_cls).items()
    }

    out = {
        field_name: resolver.execute(context=context, model=model, field_name=field_name)
        for field_name, resolver in field_resolvers.items()
        # filter out unset fields to trigger defaults
        if (resolver.model_field_name or field_name) in model.model_dump(exclude_unset=True)
        and getattr(model, resolver.model_field_name or field_name) != _Unset
    }
    return {alias_name_by_field_name[k]: v for k, v in out.items()}


T = TypeVar("T")


def _get_resolver(annotation: Any, field_name: str) -> "Resolver":
    origin = get_origin(annotation)
    args = get_args(annotation)

    # explicit field level Resolver
    if origin is Annotated:
        resolver = next((arg for arg in args if isinstance(arg, Resolver)), None)
        if resolver:
            # if the outer resolver is default, see if there is a nested one
            if resolver.is_default:
                nested = _dig_for_resolver(args[0], [])
                if nested:
                    return nested.with_outer_resolver(resolver)

            check.invariant(
                _is_resolvable_type(args[0]) or resolver.model_field_type,
                f"Resolver for {field_name} must define model_field_type, {args[0]} is not model compliant.",
            )
            return resolver

    # nested or implicit
    res = _dig_for_resolver(annotation, [])
    if res:
        return res

    from dagster.components.resolved.core_models import CORE_MODEL_SUGGESTIONS

    core_model_suggestion = ""
    if annotation in CORE_MODEL_SUGGESTIONS:
        core_model_suggestion = f"\n\nAn annotated resolver for {annotation.__name__} is available, you may wish to use it instead: {CORE_MODEL_SUGGESTIONS[annotation]}"

    raise ResolutionException(
        "Could not derive resolver for annotation\n"
        f"  {field_name}: {annotation}\n"
        "Field types are expected to be:\n"
        "* serializable types such as str, float, int, bool, list, Enum, etc\n"
        "* Resolvable subclasses\n"
        "* pydantic Models\n"
        "* Annotated with an appropriate dagster.components.Resolver\n"
        f"  e.g. Annotated[{annotation.__name__}, Resolver(fn=..., model_field_type=...)]"
        f"{core_model_suggestion}"
    )


def _dig_for_resolver(annotation, path: Sequence[_TypeContainer]) -> Resolver | None:
    if _is_implicitly_resolved_type(annotation):
        return Resolver.default()

    origin = get_origin(annotation)
    args = get_args(annotation)
    if safe_is_subclass(annotation, Resolvable):
        return Resolver(
            partial(
                _resolve_at_path,
                container_path=path,
                resolver=annotation.resolve_from_model,
            ),
            model_field_type=_wrap(annotation.model(), path),
        )

    if origin is Annotated:
        resolver = next((arg for arg in args if isinstance(arg, Resolver)), None)
        if resolver:
            check.invariant(
                _is_resolvable_type(args[0]) or resolver.model_field_type,
                f"Nested resolver must define model_field_type {args[0]} is not model compliant.",
            )
            # need to ensure nested resolvers set their model type
            if resolver.resolves_from_parent_object and path:
                raise ResolutionException(
                    f"Resolver.from_model found nested within {list(p.name for p in path)}. "
                    "Resolver.from_model can only be used on the outer most Annotated wrapper."
                )

            return Resolver(
                resolver.fn.__class__(
                    partial(
                        _resolve_at_path,
                        container_path=path,
                        resolver=resolver.fn.callable,
                    )
                ),
                model_field_type=_wrap(resolver.model_field_type or args[0], path),
                inject_before_resolve=resolver.inject_before_resolve,
            )
        annotated_type = args[0]
        if _is_implicitly_resolved_type(annotated_type):
            return Resolver.default()

        return _dig_for_resolver(annotated_type, path)

    if origin in (Union, UnionType):
        if len(args) == 2 and args[1] is type(None):
            res = _dig_for_resolver(args[0], [*path, _TypeContainer.OPTIONAL])
            if res:
                return res
        else:
            arg_resolver_pairs = [(arg, _dig_for_resolver(arg, path)) for arg in args]
            if all(r is not None for _, r in arg_resolver_pairs):
                return Resolver.union(
                    arg_resolver_pairs,  # type: ignore # doesn't understand all check
                )

    elif origin in (
        Sequence,
        tuple,
        list,
    ):  # should look for tuple[T, ...] specifically
        res = _dig_for_resolver(args[0], [*path, _TypeContainer.SEQUENCE])
        if res:
            return res

    elif origin is dict:
        key_type, value_type = args
        if key_type != str:
            raise ResolutionException(f"dict key type must be str, got {key_type}")
        value_res = _dig_for_resolver(value_type, [*path, _TypeContainer.DICT])
        if value_res:
            return value_res


def _wrap(ttype, path: Sequence[_TypeContainer]):
    result_type = ttype
    for container in reversed(path):
        if container is _TypeContainer.OPTIONAL:
            result_type = Optional[result_type]  # noqa: UP045
        elif container is _TypeContainer.SEQUENCE:
            # use tuple instead of Sequence for perf
            result_type = tuple[result_type, ...]
        elif container is _TypeContainer.DICT:
            result_type = dict[str, result_type]
        else:
            check.assert_never(container)
    return result_type


def _resolve_at_path(
    context: "ResolutionContext",
    value: Any,
    container_path: Sequence[_TypeContainer],
    resolver,
):
    if not container_path:
        return resolver(context, value)

    container = container_path[0]
    inner_path = container_path[1:]
    if container is _TypeContainer.OPTIONAL:
        return _resolve_at_path(context, value, inner_path, resolver) if value is not None else None
    elif container is _TypeContainer.SEQUENCE:
        return [
            _resolve_at_path(context.at_path(idx), i, inner_path, resolver)
            for idx, i in enumerate(value)
        ]
    elif container is _TypeContainer.DICT:
        return {
            k: _resolve_at_path(context.at_path(k), v, inner_path, resolver)
            for k, v in value.items()
        }

    check.assert_never(container)


def _ensure_non_resolvable_model_compliance(mtype: type[BaseModel]):
    for name, field_info in mtype.model_fields.items():
        field_type = field_info.rebuild_annotation()
        if not _is_implicitly_resolved_type(field_type):
            raise ResolutionException(
                f"pydantic model class {mtype.__name__} includes incompatible field\n"
                f"  {name}: {field_type}\n"
                "Subclass Resolvable to support Resolvers on fields."
            )
